from __future__ import annotations
from copy import deepcopy
from pathlib import Path
from typing import Any, Literal
from collections.abc import Sequence

import numpy as np
import pandas as pd
from tqdm import tqdm

from mikecore.DfsFactory import DfsBuilder, DfsFactory
from mikecore.DfsFile import DfsFile, DfsSimpleType
from mikecore.DfsFileFactory import DfsFileFactory
from mikecore.eum import eumQuantity, eumUnit
from mikecore.Projections import Cartography

from .. import __dfs_version__
from ..dataset import Dataset
from ._dfs import (
    _Dfs123,
    _get_item_info,
    _valid_item_numbers,
    _valid_timesteps,
    write_dfs_data,
)
from ..eum import TimeStepUnit
from ..spatial import Grid2D


def write_dfs2(filename: str | Path, ds: Dataset, title: str = "") -> None:
    dfs = _write_dfs2_header(filename, ds, title)
    write_dfs_data(dfs=dfs, ds=ds, n_spatial_dims=2)


def _write_dfs2_header(filename: str | Path, ds: Dataset, title: str = "") -> DfsFile:
    builder = DfsBuilder.Create(title, "mikeio", __dfs_version__)
    builder.SetDataType(0)

    geometry: Grid2D = ds.geometry

    if (
        geometry._shift_origin_on_write
        and not geometry._is_rotated
        and not geometry.is_spectral
    ):
        geometry = deepcopy(ds.geometry)
        geometry._shift_x0y0_to_origin()

    factory = DfsFactory()
    _write_dfs2_spatial_axis(builder, factory, geometry)
    proj_str = geometry.projection_string
    origin = geometry.origin
    orient = geometry.orientation

    if geometry.is_geo:
        proj = factory.CreateProjectionGeoOrigin(proj_str, *origin, orient)
    else:
        cart: Cartography = Cartography.CreateProjOrigin(proj_str, *origin, orient)
        proj = factory.CreateProjectionGeoOrigin(
            wktProjectionString=geometry.projection,
            lon0=cart.LonOrigin,
            lat0=cart.LatOrigin,
            orientation=cart.Orientation,
        )

    builder.SetGeographicalProjection(proj)

    timestep_unit = TimeStepUnit.SECOND
    dt = ds.timestep or 1.0  # It can not be None
    if ds.is_equidistant:
        time_axis = factory.CreateTemporalEqCalendarAxis(
            timestep_unit, ds.time[0], 0, dt
        )
    else:
        time_axis = factory.CreateTemporalNonEqCalendarAxis(timestep_unit, ds.time[0])
    builder.SetTemporalAxis(time_axis)

    for item in ds.items:
        builder.AddCreateDynamicItem(
            item.name,
            eumQuantity.Create(item.type, item.unit),
            DfsSimpleType.Float,
            item.data_value_type,
        )

    try:
        builder.CreateFile(str(filename))
    except IOError:
        print("cannot create dfs file: ", filename)

    return builder.GetFile()


def _write_dfs2_spatial_axis(
    builder: DfsBuilder, factory: DfsFactory, geometry: Grid2D
) -> None:
    builder.SetSpatialAxis(
        factory.CreateAxisEqD2(
            eumUnit.eumUmeter,
            geometry._nx,
            geometry._x0,
            geometry._dx,
            geometry._ny,
            geometry._y0,
            geometry._dy,
        )
    )


class Dfs2(_Dfs123):
    """Class for reading/writing dfs2 files."""

    _ndim = 2

    def __init__(
        self,
        filename: str | Path,
        type: Literal["horizontal", "spectral", "vertical"] = "horizontal",
    ):
        filename = str(filename)
        super().__init__(filename)

        is_spectral = type == "spectral"
        is_vertical = type == "vertical"
        dfs = DfsFileFactory.Dfs2FileOpen(str(filename))

        x0 = dfs.SpatialAxis.X0 if is_spectral else 0.0
        y0 = dfs.SpatialAxis.Y0 if is_spectral else 0.0

        origin, orientation = self._origin_and_orientation_in_CRS()

        self._geometry = Grid2D(
            dx=dfs.SpatialAxis.Dx,
            dy=dfs.SpatialAxis.Dy,
            nx=dfs.SpatialAxis.XCount,
            ny=dfs.SpatialAxis.YCount,
            x0=x0,
            y0=y0,
            projection=self._projstr,
            orientation=orientation,
            origin=origin,
            is_spectral=is_spectral,
            is_vertical=is_vertical,
        )
        dfs.Close()
        self._validate_no_orientation_in_geo()

    def read(
        self,
        *,
        items: str | int | Sequence[str | int] | None = None,
        time: int | str | slice | None = None,
        area: tuple[float, float, float, float] | None = None,
        keepdims: bool = False,
        dtype: Any = np.float32,
    ) -> Dataset:
        """Read data from a dfs2 file.

        Parameters
        ---------
        items: list[int] or list[str], optional
            Read only selected items, by number (0-based), or by name
        time: int, str, datetime, pd.TimeStamp, sequence, slice or pd.DatetimeIndex, optional
            Read only selected time steps, by default None (=all)
        keepdims: bool, optional
            When reading a single time step only, should the time-dimension be kept
            in the returned Dataset? by default: False
        area: array[float], optional
            Read only data inside (horizontal) area given as a
            bounding box (tuple with left, lower, right, upper) coordinates
        dtype: data-type, optional
            Define the dtype of the returned dataset (default = np.float32)
        Returns
        -------
        Dataset

        """
        self._open()

        item_numbers = _valid_item_numbers(self._dfs.ItemInfo, items)
        n_items = len(item_numbers)
        items = _get_item_info(self._dfs.ItemInfo, item_numbers)

        single_time_selected, time_steps = _valid_timesteps(self._dfs.FileInfo, time)
        nt = len(time_steps) if not single_time_selected else 1

        shape: tuple[int, ...]

        if area is not None:
            take_subset = True
            ii, jj = self.geometry.find_index(area=area)  # type: ignore
            shape = (nt, len(jj), len(ii))
            geometry = self.geometry._index_to_Grid2D(ii, jj)
        else:
            take_subset = False
            shape = (nt, self.ny, self.nx)
            geometry = self.geometry

        if single_time_selected and not keepdims:
            shape = shape[1:]

        data_list: list[np.ndarray] = [
            np.ndarray(shape=shape, dtype=dtype) for _ in range(n_items)
        ]

        t_seconds = np.zeros(len(time_steps))

        for i, it in enumerate(tqdm(time_steps, disable=not self.show_progress)):
            for item in range(n_items):
                itemdata = self._dfs.ReadItemTimeStep(item_numbers[item] + 1, int(it))
                d = itemdata.Data

                d[d == self.deletevalue] = np.nan
                d = d.reshape(self.ny, self.nx)

                if take_subset:
                    d = np.take(np.take(d, jj, axis=0), ii, axis=-1)

                if single_time_selected and not keepdims:
                    data_list[item] = d
                else:
                    data_list[item][i] = d

            t_seconds[i] = itemdata.Time

        self._dfs.Close()

        time = pd.to_datetime(t_seconds, unit="s", origin=self.start_time)

        dims: tuple[str, ...]

        if single_time_selected and not keepdims:
            dims = ("y", "x")
        else:
            dims = ("time", "y", "x")

        return Dataset(
            data_list,
            time=time,
            items=items,
            geometry=geometry,
            dims=dims,
            validate=False,
        )

    def append(self, ds: Dataset, validate: bool = True) -> None:
        """Append a Dataset to an existing dfs2 file.

        Parameters
        ----------
        ds: Dataset
            Dataset to append
        validate: bool, optional
            Check if the dataset to append has the same geometry and items as the original file,
            by default True

        Notes
        -----
        The original file is modified.

        """
        if validate:
            if self.geometry != ds.geometry:
                raise ValueError("The geometry of the dataset to append does not match")

            for item_s, item_o in zip(ds.items, self.items):
                if item_s != item_o:
                    raise ValueError(
                        f"Item in dataset {item_s.name} does not match {item_o.name}"
                    )

        dfs = DfsFileFactory.Dfs2FileOpenAppend(str(self._filename))
        write_dfs_data(dfs=dfs, ds=ds, n_spatial_dims=2)

        self._n_timesteps = dfs.FileInfo.TimeAxis.NumberOfTimeSteps

    def _open(self) -> None:
        self._dfs = DfsFileFactory.Dfs2FileOpen(self._filename)
        self._source = self._dfs

    @property
    def geometry(self) -> Grid2D:
        """Spatial information."""
        assert isinstance(self._geometry, Grid2D)
        return self._geometry

    @property
    def x0(self) -> Any:
        """Start point of x values (often 0)."""
        return self.geometry.x[0]

    @property
    def y0(self) -> Any:
        """Start point of y values (often 0)."""
        return self.geometry.y[0]

    @property
    def dx(self) -> float:
        """Step size in x direction."""
        return self.geometry.dx

    @property
    def dy(self) -> float:
        """Step size in y direction."""
        return self.geometry.dy

    @property
    def shape(self) -> tuple[int, ...]:
        """Tuple with number of values in the t-, y-, x-direction."""
        return (self._n_timesteps, self.geometry.ny, self.geometry.nx)

    @property
    def nx(self) -> int:
        """Number of values in the x-direction."""
        return self.geometry.nx

    @property
    def ny(self) -> int:
        """Number of values in the y-direction."""
        return self.geometry.ny



from ._dataarray import DataArray
from ._dataset import Dataset, from_pandas, from_polars
from ._data_plot import (
    _DataArrayPlotter,
    _DataArrayPlotterGrid1D,
    _DataArrayPlotterGrid2D,
    _DataArrayPlotterFM,
    _DataArrayPlotterFMVerticalColumn,
    _DataArrayPlotterFMVerticalProfile,
    _DataArrayPlotterPointSpectrum,
    _DataArrayPlotterLineSpectrum,
    _DataArrayPlotterAreaSpectrum,
    _DatasetPlotter,
)

__all__ = [
    "DataArray",
    "Dataset",
    "from_pandas",
    "from_polars",
    "_DataArrayPlotter",
    "_DataArrayPlotterGrid1D",
    "_DataArrayPlotterGrid2D",
    "_DataArrayPlotterFM",
    "_DataArrayPlotterFMVerticalColumn",
    "_DataArrayPlotterFMVerticalProfile",
    "_DataArrayPlotterPointSpectrum",
    "_DataArrayPlotterLineSpectrum",
    "_DataArrayPlotterAreaSpectrum",
    "_DatasetPlotter",
]



from __future__ import annotations
from functools import cached_property
from pathlib import Path
from typing import (
    Any,
    Literal,
    Sequence,
    Sized,
    TYPE_CHECKING,
)
import warnings


import numpy as np
from mikecore.DfsuFile import DfsuFileType
from mikecore.eum import eumQuantity
from mikecore.MeshBuilder import MeshBuilder
from scipy.spatial import cKDTree

from ..eum import EUMType, EUMUnit
from ..exceptions import OutsideModelDomainError
from .._interpolation import get_idw_interpolant, interp2d
from ._FM_utils import (
    _get_node_centered_data,
    _plot_map,
    BoundaryPolygons,
    Polygon,
    _set_xy_label_by_projection,  # TODO remove
    _to_polygons,  # TODO remove
)
from ._geometry import GeometryPoint2D, _Geometry

from ._grid_geometry import Grid2D
from ._utils import xy_to_bbox


if TYPE_CHECKING:
    from ._FM_geometry_layered import GeometryFM3D
    from matplotlib.axes import Axes


class _GeometryFMPlotter:
    """Plot GeometryFM.

    Examples
    --------
    ```{python}
    import mikeio

    ds = mikeio.read("../data/FakeLake.dfsu")
    g = ds.geometry
    g.plot()
    ```

    """

    def __init__(self, geometry: GeometryFM2D | GeometryFM3D) -> None:
        self.g = geometry

    def __call__(
        self,
        ax: Axes | None = None,
        figsize: tuple[float, float] | None = None,
        **kwargs: Any,
    ) -> Axes:
        """Plot bathymetry as coloured patches."""
        ax = self._get_ax(ax, figsize)
        kwargs["plot_type"] = kwargs.get("plot_type") or "patch"
        return self._plot_FM_map(ax, **kwargs)

    def contour(
        self,
        ax: Axes | None = None,
        figsize: tuple[float, float] | None = None,
        **kwargs: Any,
    ) -> Axes:
        """Plot bathymetry as contour lines.

        Examples
        --------
        ```{python}
        g.plot.contour()
        ```

        """
        ax = self._get_ax(ax, figsize)
        kwargs["plot_type"] = "contour"
        return self._plot_FM_map(ax, **kwargs)

    def contourf(
        self,
        ax: Axes | None = None,
        figsize: tuple[float, float] | None = None,
        **kwargs: Any,
    ) -> Axes:
        """Plot bathymetry as filled contours.

        Examples
        --------
        ```{python}
        g.plot.contourf()
        ```

        """
        ax = self._get_ax(ax, figsize)
        kwargs["plot_type"] = "contourf"
        return self._plot_FM_map(ax, **kwargs)

    @staticmethod
    def _get_ax(
        ax: Axes | None = None,
        figsize: tuple[float, float] | None = None,
    ) -> Axes:
        import matplotlib.pyplot as plt

        if ax is None:
            _, ax = plt.subplots(figsize=figsize)
        return ax

    def _plot_FM_map(self, ax: Axes, **kwargs: Any) -> Axes:
        if "title" not in kwargs:
            kwargs["title"] = "Bathymetry"

        plot_type = kwargs.pop("plot_type")

        g = self.g

        return _plot_map(
            node_coordinates=g.node_coordinates,
            element_table=g.element_table,
            element_coordinates=g.element_coordinates,
            boundary_polylines=g.boundary_polygons.lines,
            plot_type=plot_type,
            projection=g.projection,
            z=None,
            ax=ax,
            **kwargs,
        )

    def mesh(
        self,
        title: str = "Mesh",
        figsize: tuple[float, float] | None = None,
        ax: Axes | None = None,
    ) -> Axes:
        """Plot mesh only.

        Examples
        --------
        ```{python}
        g.plot.mesh()
        ```

        """
        # TODO this must be a duplicate, delegate

        from matplotlib.collections import PatchCollection  # type: ignore

        ax = self._get_ax(ax=ax, figsize=figsize)
        ax.set_aspect(self._plot_aspect())

        patches = _to_polygons(self.g.node_coordinates, self.g.element_table)
        fig_obj = PatchCollection(
            patches, edgecolor="0.6", facecolor="none", linewidths=0.3
        )
        ax.add_collection(fig_obj)
        self.outline(ax=ax)
        ax.set_title(title)
        ax = self._set_plot_limits(ax)
        _set_xy_label_by_projection(ax, self.g.projection)
        return ax

    def outline(
        self,
        title: str = "Outline",
        figsize: tuple[float, float] | None = None,
        ax: Axes | None = None,
    ) -> Axes:
        """Plot domain outline.

        Examples
        --------
        ```{python}
        g.plot.outline()
        ```

        """
        ax = self._get_ax(ax=ax, figsize=figsize)
        ax.set_aspect(self._plot_aspect())

        linwid = 1.2
        out_col = "0.4"
        for exterior in self.g.boundary_polygons.exteriors:
            ax.plot(*exterior.xy.T, color=out_col, linewidth=linwid)
        for interior in self.g.boundary_polygons.interiors:
            ax.plot(*interior.xy.T, color=out_col, linewidth=linwid)
        if title is not None:
            ax.set_title(title)
        ax = self._set_plot_limits(ax)
        return ax

    def boundary_nodes(
        self,
        boundary_names: Sequence[str] | None = None,
        figsize: tuple[float, float] | None = None,
        ax: Axes | None = None,
    ) -> Axes:
        """Plot mesh boundary nodes and their code values.

        Examples
        --------
        ```{python}
        g.plot.boundary_nodes()
        ```

        """
        import matplotlib.pyplot as plt

        ax = self._get_ax(ax=ax, figsize=figsize)
        ax.set_aspect(self._plot_aspect())

        nc = self.g.node_coordinates
        c = self.g.codes
        valid_codes = list(set(self.g.codes))
        boundary_codes = [code for code in valid_codes if code > 0]

        if boundary_names is not None:
            if len(boundary_codes) != len(boundary_names):
                raise Exception(
                    f"Number of boundary names ({len(boundary_names)}) inconsistent with number of boundaries ({len(boundary_codes)})"
                )
            user_defined_labels = dict(zip(boundary_codes, boundary_names))

        for code in boundary_codes:
            xn = nc[c == code, 0]
            yn = nc[c == code, 1]
            if boundary_names is None:
                label = f"Code {code}"
            else:
                label = user_defined_labels[code]
            plt.plot(xn, yn, ".", label=label)

        plt.legend()
        plt.title("Boundary nodes")
        ax = self._set_plot_limits(ax)
        return ax

    def _set_plot_limits(self, ax: Axes) -> Axes:
        bbox = xy_to_bbox(self.g.node_coordinates)
        xybuf = 6e-3 * (bbox.right - bbox.left)
        ax.set_xlim(bbox.left - xybuf, bbox.right + xybuf)
        ax.set_ylim(bbox.bottom - xybuf, bbox.top + xybuf)
        return ax

    def _plot_aspect(self) -> Literal["equal"] | float:
        if self.g.is_geo:
            mean_lat = np.mean(self.g.node_coordinates[:, 1])
            return 1.0 / np.cos(np.pi * mean_lat / 180)
        else:
            return "equal"


class _GeometryFM(_Geometry):
    def __init__(
        self,
        *,
        node_coordinates: np.ndarray,
        element_table: np.ndarray | list[Sequence[int]] | list[np.ndarray],
        projection: str,
        codes: np.ndarray | None = None,
        dfsu_type: DfsuFileType,
        element_ids: np.ndarray | None = None,
        node_ids: np.ndarray | None = None,
        validate: bool = True,
        reindex: bool = False,
    ) -> None:
        super().__init__(projection=projection)
        self.node_coordinates = np.asarray(node_coordinates)

        n_nodes = len(node_coordinates)
        self._codes = (
            np.zeros((n_nodes,), dtype=int) if codes is None else np.asarray(codes)
        )

        self._node_ids = (
            np.arange(len(self._codes)) if node_ids is None else np.asarray(node_ids)
        )

        self._type = dfsu_type

        self.element_table, self._element_ids = self._check_elements(
            element_table=element_table,
            element_ids=element_ids,
            validate=validate,
        )

        if reindex:
            self._reindex()

    def _calc_element_coordinates(self, maxnodes: int = 4) -> np.ndarray:
        element_table = self.element_table

        n_elements = len(element_table)
        ec = np.empty([n_elements, 3])

        # pre-allocate for speed
        idx = np.zeros(maxnodes, dtype=int)
        xcoords = np.zeros([maxnodes, n_elements])
        ycoords = np.zeros([maxnodes, n_elements])
        zcoords = np.zeros([maxnodes, n_elements])
        nnodes_per_elem = np.zeros(n_elements)

        for j in range(n_elements):
            nodes = element_table[j]
            nnodes = len(nodes)
            nnodes_per_elem[j] = nnodes
            for i in range(nnodes):
                idx[i] = nodes[i]  # - 1

            x, y, z = self.node_coordinates[idx[:nnodes]].T

            xcoords[:nnodes, j] = x
            ycoords[:nnodes, j] = y
            zcoords[:nnodes, j] = z

        ec[:, 0] = np.sum(xcoords, axis=0) / nnodes_per_elem
        ec[:, 1] = np.sum(ycoords, axis=0) / nnodes_per_elem
        ec[:, 2] = np.sum(zcoords, axis=0) / nnodes_per_elem

        return ec

    def _check_elements(
        self,
        element_table: np.ndarray | list[Sequence[int]] | list[np.ndarray],
        element_ids: np.ndarray | None = None,
        validate: bool = True,
    ) -> tuple[Any, Any]:
        if validate:
            max_node_id = self._node_ids.max()
            for i, e in enumerate(element_table):
                # TODO: avoid looping through all elements (could be +1e6)!
                if not isinstance(e, np.ndarray):
                    e = np.asarray(e)
                    element_table[i] = e

                # NOTE: this check "e.max()" takes the most of the time when constructing a new FM_geometry
                if e.max() > max_node_id:
                    raise ValueError(
                        f"Element table has node # {e.max()}. Max node id: {max_node_id}"
                    )

        if element_ids is None:
            element_ids = np.arange(len(element_table))
        element_ids = np.asarray(element_ids)

        # TODO make sure return type is np.ndarray
        return element_table, element_ids

    def _reindex(self) -> None:
        new_node_ids = np.arange(self.n_nodes)
        new_element_ids = np.arange(self.n_elements)
        node_dict = dict(zip(self._node_ids, new_node_ids))
        for eid in range(self.n_elements):
            elem_nodes = self.element_table[eid]
            new_elem_nodes = np.zeros_like(elem_nodes)
            for jn, idx in enumerate(elem_nodes):
                new_elem_nodes[jn] = node_dict[idx]
            self.element_table[eid] = new_elem_nodes

        self._node_ids = new_node_ids
        self._element_ids = new_element_ids

    @property
    def default_dims(self) -> tuple[str, ...]:
        return ("element",)

    @property
    def is_spectral(self) -> bool:
        return False

    @property
    def n_nodes(self) -> int:
        """Number of nodes."""
        return len(self._node_ids)

    @property
    def node_ids(self) -> np.ndarray:
        return self._node_ids

    @property
    def n_elements(self) -> int:
        """Number of elements."""
        return len(self._element_ids)

    @property
    def element_ids(self) -> np.ndarray:
        return self._element_ids

    @cached_property
    def max_nodes_per_element(self) -> int:
        """The maximum number of nodes for an element."""
        maxnodes = 0
        for local_nodes in self.element_table:
            n = len(local_nodes)
            if n > maxnodes:
                maxnodes = n
        return maxnodes

    @property
    def codes(self) -> np.ndarray:
        """Node codes of all nodes (0=water, 1=land, 2...=open boundaries)."""
        return self._codes

    @codes.setter
    def codes(self, v: np.ndarray) -> None:
        if len(v) != self.n_nodes:
            raise ValueError(f"codes must have length of nodes ({self.n_nodes})")
        self._codes = np.array(v, dtype=np.int32)

    @property
    def boundary_codes(self) -> list[int]:
        """Unique list of boundary codes."""
        valid = list(set(self.codes))
        return [code for code in valid if code > 0]

    def __eq__(self, value: Any) -> bool:
        # this is not an exhaustive check, but should be sufficient for most cases

        if self.__class__ != value.__class__:
            return False

        if self.projection != value.projection:
            return False

        if not np.array_equal(self.node_coordinates, value.node_coordinates):
            return False

        return True


class GeometryFM2D(_GeometryFM):
    """Flexible 2d mesh geometry."""

    def __init__(
        self,
        node_coordinates: np.ndarray,
        # TODO settle on type for element_table
        element_table: Any,
        codes: np.ndarray | None = None,
        projection: str = "LONG/LAT",
        dfsu_type: DfsuFileType = DfsuFileType.Dfsu2D,  # Reasonable default?
        element_ids: np.ndarray | None = None,
        node_ids: np.ndarray | None = None,
        validate: bool = True,
        reindex: bool = False,
    ) -> None:
        super().__init__(
            node_coordinates=node_coordinates,
            element_table=element_table,  # type: ignore
            codes=codes,
            projection=projection,
            dfsu_type=dfsu_type,
            element_ids=element_ids,
            node_ids=node_ids,
            validate=validate,
            reindex=reindex,
        )

        self.plot = _GeometryFMPlotter(self)

    def __str__(self) -> str:
        return f"{self.type_name} ({self.n_elements} elements, {self.n_nodes} nodes)"

    def __repr__(self) -> str:
        return (
            f"Flexible Mesh Geometry: {self._type.name}\n"
            f"number of nodes: {self.n_nodes}\n"
            f"number of elements: {self.n_elements}\n"
            f"projection: {self.projection_string}"
        )

    @staticmethod
    def _point_in_polygon(xn: np.ndarray, yn: np.ndarray, xp: float, yp: float) -> bool:
        """Check for each side in the polygon that the point is on the correct side."""
        for j in range(len(xn) - 1):
            if (yn[j + 1] - yn[j]) * (xp - xn[j]) + (-xn[j + 1] + xn[j]) * (
                yp - yn[j]
            ) > 0:
                return False
            if (yn[0] - yn[-1]) * (xp - xn[-1]) + (-xn[0] + xn[-1]) * (yp - yn[-1]) > 0:
                return False
        return True

    @staticmethod
    def _area_is_bbox(area: Sized) -> bool:
        return isinstance(area, Sized) and len(area) == 4

    @staticmethod
    def _area_is_polygon(area: Sequence[tuple[float, float]] | Sequence[float]) -> bool:
        polygon = np.array(area)
        return polygon.ndim == 2 and polygon.shape[1] == 2

    @property
    def type_name(self) -> str:
        """Type name, e.g. Mesh, Dfsu2D."""
        return self._type.name if self._type else "Mesh"

    @property
    def ndim(self) -> int:
        return 2

    @property
    def geometry2d(self) -> GeometryFM2D:
        return self

    @property
    def is_2d(self) -> bool:
        """Type is either mesh or Dfsu2D (2 horizontal dimensions)."""
        return self._type in (
            DfsuFileType.Dfsu2D,
            DfsuFileType.DfsuSpectral2D,
            None,
        )

    @property
    def is_layered(self) -> bool:
        """Type is layered dfsu (3d, vertical profile or vertical column)."""
        return False

    @property
    def is_spectral(self) -> bool:
        """Type is spectral dfsu (point, line or area spectrum)."""
        return self._type in (
            DfsuFileType.DfsuSpectral0D,
            DfsuFileType.DfsuSpectral1D,
            DfsuFileType.DfsuSpectral2D,
        )

    @property
    def is_tri_only(self) -> bool:
        """Does the mesh consist of triangles only."""
        return self.max_nodes_per_element == 3 or self.max_nodes_per_element == 6

    @cached_property
    def element_coordinates(self) -> np.ndarray:
        """Center coordinates of each element."""
        return self._calc_element_coordinates()

    @cached_property
    def _tree2d(self) -> cKDTree:
        xy = self.element_coordinates[:, :2]
        return cKDTree(xy)

    def find_nearest_elements(
        self,
        x: float | np.ndarray,
        y: float | np.ndarray | None = None,
        n_nearest: int = 1,
        return_distances: bool = False,
    ) -> Any:
        """Find index of nearest elements (optionally for a list).

        Parameters
        ----------
        x: float or array(float)
            X coordinate(s) (easting or longitude)
        y: float or array(float)
            Y coordinate(s) (northing or latitude)
        n_nearest : int, optional
            return this many (horizontally) nearest points for
            each coordinate set, default=1
        return_distances : bool, optional
            should the horizontal distances to each point be returned?
            default=False

        Returns
        -------
        np.array
            element ids of nearest element(s)
        np.array, optional
            horizontal distances

        Examples
        --------
        >>> g = dfs.geometry
        >>> id = g.find_nearest_elements(3, 4)
        >>> ids = g.find_nearest_elements([3, 8], [4, 6])
        >>> ids = g.find_nearest_elements(xy)
        >>> ids = g.find_nearest_elements(3, 4, n_nearest=4)
        >>> ids, d = g.find_nearest_elements(xy, return_distances=True)


        See Also
        --------
        find_index : find element indicies for points or an area

        """
        idx, d2d = self._find_n_nearest_2d_elements(x, y, n=n_nearest)

        if return_distances:
            return idx, d2d

        return idx

    def get_2d_interpolant(
        self,
        xy: np.ndarray,
        n_nearest: int = 5,
        extrapolate: bool = False,
        p: int = 2,
        radius: float | None = None,
    ) -> tuple[Any, Any]:
        """IDW interpolant for list of coordinates.

        Parameters
        ----------
        xy : array-like
            x,y coordinates of new points
        n_nearest : int, optional
            number of nearest elements used for IDW, by default 5
        extrapolate : bool, optional
            allow extrapolation, by default False
        p : float, optional
            power of inverse distance weighting, default=2
        radius: float, optional
            an alternative to extrapolate=False,
            only include elements within radius

        Returns
        -------
        (np.array, np.array)
            element ids and weights

        """
        xy = np.atleast_2d(xy)
        ids, dists = self._find_n_nearest_2d_elements(xy, n=n_nearest)
        weights = None

        if n_nearest == 1:
            weights = np.ones(dists.shape)
            if not extrapolate:
                weights[~self.contains(xy)] = np.nan  # type: ignore
        elif n_nearest > 1:
            weights = get_idw_interpolant(dists, p=p)
            if not extrapolate:
                weights[~self.contains(xy), :] = np.nan  # type: ignore
        else:
            ValueError("n_nearest must be at least 1")

        if radius is not None:
            weights[dists > radius] = np.nan  # type: ignore

        return ids, weights

    def interp2d(
        self,
        data: np.ndarray,
        elem_ids: np.ndarray,
        weights: np.ndarray | None = None,
        shape: tuple[int, ...] | None = None,
    ) -> np.ndarray | list[np.ndarray]:
        """Interpolate spatially in data (2d only).

        Parameters
        ----------
        data : ndarray or list(ndarray)
            dfsu data
        elem_ids : ndarray(int)
            n sized array of 1 or more element ids used for interpolation
        weights : ndarray(float), optional
            weights with same size as elem_ids used for interpolation
        shape: tuple, optional
            reshape output

        Returns
        -------
        ndarray or list(ndarray)
            spatially interped data

        Examples
        --------
        >>> ds = dfsu.read()
        >>> g = dfs.get_overset_grid(shape=(50,40))
        >>> elem_ids, weights = dfs.get_2d_interpolant(g.xy)
        >>> dsi = dfs.interp2d(ds, elem_ids, weights)

        """
        return interp2d(data, elem_ids, weights, shape)  # type: ignore

    def _find_n_nearest_2d_elements(
        self, x: float | np.ndarray, y: float | np.ndarray | None = None, n: int = 1
    ) -> tuple[Any, Any]:
        # TODO return arguments in the same order than cKDTree.query?

        if n > self.n_elements:
            raise ValueError(
                f"Cannot find {n} nearest! Number of elements: {self.n_elements}"
            )

        if y is None:
            p = x
            if (not np.isscalar(x)) and (np.ndim(x) == 2):
                p = x[:, 0:2]  # type: ignore
        else:
            p = np.array((x, y)).T
        d, elem_id = self._tree2d.query(p, k=n)
        return elem_id, d

    def _find_element_2d(self, coords: np.ndarray) -> Any:
        points_outside = []

        coords = np.atleast_2d(coords)
        nc = self.node_coordinates

        few_nearest, _ = self._find_n_nearest_2d_elements(
            coords, n=min(self.n_elements, 2)
        )
        ids = np.atleast_2d(few_nearest)[:, 0]  # first guess

        for k in range(len(ids)):
            # step 1: is nearest element = element containing point?
            nodes = self.element_table[ids[k]]
            element_found = self._point_in_polygon(
                nc[nodes, 0], nc[nodes, 1], coords[k, 0], coords[k, 1]
            )

            # step 2: if not, then try second nearest point
            if not element_found and self.n_elements > 1:
                candidate = few_nearest[k, 1]
                assert np.isscalar(candidate)
                nodes = self.element_table[candidate]
                element_found = self._point_in_polygon(
                    nc[nodes, 0], nc[nodes, 1], coords[k, 0], coords[k, 1]
                )
                ids[k] = few_nearest[k, 1]

            # step 3: if not, then try with *many* more points
            if not element_found and self.n_elements > 1:
                many_nearest, _ = self._find_n_nearest_2d_elements(
                    coords[k, :],
                    n=min(self.n_elements, 10),  # TODO is 10 enough?
                )
                for p in many_nearest[2:]:  # we have already tried the two first above
                    nodes = self.element_table[p]
                    element_found = self._point_in_polygon(
                        nc[nodes, 0], nc[nodes, 1], coords[k, 0], coords[k, 1]
                    )
                    if element_found:
                        ids[k] = p
                        break

            if not element_found:
                points_outside.append(k)

        if len(points_outside) > 0:
            raise OutsideModelDomainError(  # type: ignore
                x=coords[points_outside, 0],
                y=coords[points_outside, 1],
                indices=points_outside,
            )

        return ids

    def get_overset_grid(
        self,
        dx: float | None = None,
        dy: float | None = None,
        nx: int | None = None,
        ny: int | None = None,
        buffer: float = 0.0,
    ) -> Grid2D:
        """Get a 2d grid that covers the domain by specifying spacing or shape.

        Parameters
        ----------
        dx : float or (float, float), optional
            grid resolution in x-direction (or in x- and y-direction)
        dy : float, optional
            grid resolution in y-direction
        nx : int, optional
            number of points in x-direction, by default None,
            (the value will be inferred)
        ny : int, optional
            number of points in y-direction, by default None,
            (the value will be inferred)
        buffer : float, optional
            positive to make the area larger, default=0
            can be set to a small negative value to avoid NaN
            values all around the domain.

        Returns
        -------
        <mikeio.Grid2D>
            2d grid

        """
        nc = self.node_coordinates
        bbox = xy_to_bbox(nc, buffer=buffer)
        return Grid2D(bbox=bbox, dx=dx, dy=dy, nx=nx, ny=ny, projection=self.projection)

    def get_element_area(self) -> np.ndarray:
        """Calculate the horizontal area of each element.

        Returns
        -------
        np.array(float)
            areas in m2

        """
        n_elements = self.n_elements

        # Node coordinates
        xn = self.node_coordinates[:, 0]
        yn = self.node_coordinates[:, 1]

        area = np.empty(n_elements)
        xcoords = np.empty(8)
        ycoords = np.empty(8)

        for j in range(n_elements):
            nodes = self.element_table[j]
            n_nodes = len(nodes)

            for i in range(n_nodes):
                nidx = nodes[i]
                xcoords[i] = xn[nidx]
                ycoords[i] = yn[nidx]

            # ab : edge vector corner a to b
            abx = xcoords[1] - xcoords[0]
            aby = ycoords[1] - ycoords[0]

            # ac : edge vector corner a to c
            acx = xcoords[2] - xcoords[0]
            acy = ycoords[2] - ycoords[0]

            isquad = False
            if n_nodes > 3:
                isquad = True
                # ad : edge vector corner a to d
                adx = xcoords[3] - xcoords[0]
                ady = ycoords[3] - ycoords[0]

            # if geographical coords, convert all length to meters
            if self.is_geo:
                earth_radius = 6366707.0
                deg_to_rad = np.pi / 180.0
                earth_radius_deg_to_rad = earth_radius * deg_to_rad

                # Y on element centers
                Ye = np.sum(ycoords[:n_nodes]) / n_nodes
                cosYe = np.cos(np.deg2rad(Ye))

                abx = earth_radius_deg_to_rad * abx * cosYe
                aby = earth_radius_deg_to_rad * aby
                acx = earth_radius_deg_to_rad * acx * cosYe
                acy = earth_radius_deg_to_rad * acy
                if isquad:
                    adx = earth_radius_deg_to_rad * adx * cosYe
                    ady = earth_radius_deg_to_rad * ady

            # calculate area in m2
            area[j] = 0.5 * (abx * acy - aby * acx)
            if isquad:
                area[j] = area[j] + 0.5 * (acx * ady - acy * adx)

        return np.abs(area)

    @cached_property
    def boundary_polylines(self) -> BoundaryPolygons:
        warnings.warn(
            "boundary_polylines is renamed to boundary_polygons", FutureWarning
        )
        return self._get_boundary_polygons()

    @cached_property
    def boundary_polygons(self) -> BoundaryPolygons:
        """Lists of polygons defining domain outline."""
        return self._get_boundary_polygons()

    def contains(self, points: np.ndarray) -> np.ndarray:
        """Test if a list of points are contained by mesh.

        Parameters
        ----------
        points : array-like n-by-2
            x,y-coordinates of n points to be tested

        Returns
        -------
        bool array
            True for points inside, False otherwise

        """
        points = np.atleast_2d(points)

        return self.boundary_polygons.contains(points)

    def __contains__(self, pt: np.ndarray) -> bool:
        return self.contains(pt)[0]

    def _get_boundary_polygons_uncategorized(self) -> list[list[np.int64]]:
        """Construct closed polygons for all boundary faces."""
        boundary_faces = self._get_boundary_faces()
        face_remains = boundary_faces.copy()
        polygons = []
        while face_remains.shape[0] > 1:
            n0 = face_remains[:, 0]
            n1 = face_remains[:, 1]
            polyline = [n0[0], n1[0]]
            index_to_delete = [0]
            count = 0
            end_points = face_remains[0, 1]
            while True:
                next_point_index = np.where(n0 == end_points)
                if next_point_index[0].size != 0:
                    polyline.append(face_remains[next_point_index[0][0], 1])
                    index_to_delete.append(next_point_index[0][0])
                    end_points = polyline[-1]
                count += 1
                if count > face_remains.shape[0] or polyline[0] == end_points:
                    break

            face_remains = np.delete(face_remains, index_to_delete, axis=0)
            polygons.append(polyline)
        return polygons

    def _get_boundary_polygons(self) -> BoundaryPolygons:
        """Get boundary polylines and categorize as inner or outer by assessing the signed area."""
        polygons = self._get_boundary_polygons_uncategorized()

        interiors = []
        exteriors = []

        for polygon in polygons:
            polygon_np = np.asarray(polygon)
            xy = self.node_coordinates[polygon_np, 0:2]
            poly = Polygon(xy)
            if poly.area > 0:
                exteriors.append(poly)
            else:
                interiors.append(poly)

        return BoundaryPolygons(exteriors=exteriors, interiors=interiors)

    def _get_boundary_faces(self) -> np.ndarray:
        """Construct list of faces."""
        element_table = self.element_table

        all_faces = []
        for el in element_table:
            ele = [*el, el[0]]
            for j in range(len(el)):
                all_faces.append(ele[j : j + 2])

        all_faces = np.asarray(all_faces)

        all_faces_sorted = np.sort(all_faces, axis=1)
        _, uf_id, face_counts = np.unique(
            all_faces_sorted, axis=0, return_index=True, return_counts=True
        )

        # boundary faces are those appearing only once
        bnd_face_id = face_counts == 1
        return all_faces[uf_id[bnd_face_id]]

    def isel(
        self, idx: Sequence[int], keepdims: bool = False, **kwargs: Any
    ) -> "GeometryFM2D" | GeometryPoint2D:
        """Export a selection of elements to a new geometry.

        Typically not called directly, but by Dataset/DataArray's
        isel() or sel() methods.

        Parameters
        ----------
        idx : list(int)
            collection of element indicies
        keepdims : bool, optional
            Should the original Geometry type be kept (keepdims=True)
            or should it be reduced e.g. to a GeometryPoint2D if possible
            (keepdims=False), by default False
        **kwargs: Any
            Not used

        Returns
        -------
        Geometry
            geometry subset

        See Also
        --------
        find_index : find element indicies for points or an area

        """
        return self.elements_to_geometry(elements=idx, keepdims=keepdims)

    def find_index(
        self,
        x: float | np.ndarray | None = None,
        y: float | np.ndarray | None = None,
        coords: np.ndarray | None = None,
        area: (
            tuple[float, float, float, float] | Sequence[tuple[float, float]] | None
        ) = None,
    ) -> np.ndarray:
        """Find a *set* of element indicies for a number of points or within an area.

        The returned indices returned are the unique, unordered set of element indices that contain the points or area.

        This method will return elements *containing* the argument
        points/area, which is not necessarily the same as the nearest.

        Typically not called directly, but by Dataset/DataArray's
        sel() method.

        Parameters
        ----------
        x: float or array(float)
            X coordinate(s) (easting or longitude)
        y: float or array(float)
            Y coordinate(s) (northing or latitude)
        coords : np.array(float,float), optional
            As an alternative to specifying x, and y individually,
            the argument coords can be used instead.
            (x,y)-coordinates of points to be found,
            by default None
        area : (float, float, float, float), optional
            Bounding box of coordinates (left lower and right upper)
            to be selected, by default None

        Returns
        -------
        np.array
            indicies of containing elements

        Raises
        ------
        ValueError
            if any point is outside the domain

        Examples
        --------
        >>> g = dfs.geometry
        >>> id = dfs.find_index(x=3.1, y=4.3)

        See Also
        --------
        isel : get subset geometry for specific indicies
        find_nearest_elements : find nearest instead of containing elements

        """
        if (coords is not None) or (x is not None) or (y is not None):
            if area is not None:
                raise ValueError(
                    "Coordinates and area cannot be provided at the same time!"
                )
            if coords is not None:
                coords = np.atleast_2d(coords)
                xy = coords[:, :2]  # type: ignore
            else:
                xy = np.vstack((x, y)).T  # type: ignore
            idx = self._find_element_2d(coords=xy)
            return idx
        elif area is not None:
            return self._elements_in_area(area)
        else:
            raise ValueError("Provide either coordinates or area")

    @staticmethod
    def _inside_polygon(polygon: np.ndarray, xy: np.ndarray) -> np.ndarray:
        import matplotlib.path as mp

        if polygon.ndim == 1:
            polygon = np.column_stack((polygon[0::2], polygon[1::2]))
        return mp.Path(polygon).contains_points(xy)

    def _elements_in_area(
        self, area: Sequence[float] | Sequence[tuple[float, float]]
    ) -> np.ndarray:
        """Find 2d element ids of elements inside area."""
        if self._area_is_bbox(area):
            x0, y0, x1, y1 = area
            xc = self.element_coordinates[:, 0]
            yc = self.element_coordinates[:, 1]
            mask = (xc >= x0) & (xc <= x1) & (yc >= y0) & (yc <= y1)
        elif self._area_is_polygon(area):
            polygon = np.array(area)
            xy = self.element_coordinates[:, :2]
            mask = self._inside_polygon(polygon, xy)
        else:
            raise ValueError("'area' must be bbox [x0,y0,x1,y1] or polygon")
        elements = np.where(mask)[0]
        if len(elements) == 0:
            raise ValueError("No elements in selection!")
        return elements

    def elements_to_geometry(
        self, elements: int | Sequence[int], keepdims: bool = False
    ) -> "GeometryFM2D" | GeometryPoint2D:
        if isinstance(elements, (int, np.integer)):
            sel_elements: list[int] = [elements]
        else:
            sel_elements = list(elements)
        if len(sel_elements) == 1 and not keepdims:
            x, y, _ = self.element_coordinates[sel_elements.pop(), :]

            return GeometryPoint2D(x=x, y=y, projection=self.projection)

        # extract information for selected elements

        node_ids, elem_tbl = self._get_nodes_and_table_for_elements(sel_elements)
        node_coords = self.node_coordinates[node_ids]
        codes = self.codes[node_ids]
        elem_ids = self.element_ids[sel_elements]

        return GeometryFM2D(
            node_coordinates=node_coords,
            codes=codes,
            node_ids=node_ids,
            projection=self.projection_string,
            element_table=elem_tbl,
            element_ids=elem_ids,
            dfsu_type=self._type,
            reindex=True,
        )

    def _get_nodes_and_table_for_elements(
        self, elements: np.ndarray | list[int]
    ) -> tuple[Any, Any]:
        """List of nodes and element table for a list of elements.

        Parameters
        ----------
        elements : np.array(int)
            array of element ids

        Returns
        -------
        np.array(int)
            array of node ids (unique)
        list(list(int))
            element table with a list of nodes for each element

        """
        elem_tbl = np.empty(len(elements), dtype=np.dtype("O"))

        for j, eid in enumerate(elements):
            elem_tbl[j] = np.asarray(self.element_table[eid])

        nodes = np.unique(np.hstack(elem_tbl))  # type: ignore
        return nodes, elem_tbl

    def get_node_centered_data(
        self, data: np.ndarray, extrapolate: bool = True
    ) -> np.ndarray:
        """Convert cell-centered data to node-centered by pseudo-laplacian method.

        Parameters
        ----------
        data : np.array(float)
            cell-centered data
        extrapolate : bool, optional
            allow the method to extrapolate, default:True

        Returns
        -------
        np.array(float)
            node-centered data

        """
        geometry = self
        nc = geometry.node_coordinates
        ec = geometry.element_coordinates
        elem_table = geometry.element_table
        return _get_node_centered_data(nc, elem_table, ec, data, extrapolate)

    def to_shapely(self) -> Any:
        """Export mesh as shapely MultiPolygon.

        Returns
        -------
        shapely.geometry.MultiPolygon
            polygons with mesh elements

        """
        from shapely.geometry import MultiPolygon, Polygon  # type: ignore

        polygons = []
        for j in range(self.n_elements):
            nodes = self.element_table[j]
            pcoords = np.empty([len(nodes), 2])
            for i in range(len(nodes)):
                nidx = nodes[i]
                pcoords[i, :] = self.node_coordinates[nidx, 0:2]
            polygon = Polygon(pcoords)
            polygons.append(polygon)
        mp = MultiPolygon(polygons)

        return mp

    def to_mesh(self, outfilename: str | Path) -> None:
        """Export geometry to new mesh file.

        Parameters
        ----------
        outfilename : str
            path to file to be written

        """
        builder = MeshBuilder()
        outfilename = str(outfilename)

        nc = self.node_coordinates
        builder.SetNodes(nc[:, 0], nc[:, 1], nc[:, 2], self.codes)
        # builder.SetNodeIds(geom2d.node_ids+1)
        # builder.SetElementIds(geom2d.elements+1)
        element_table_MZ = [np.asarray(row) + 1 for row in self.element_table]
        builder.SetElements(element_table_MZ)
        builder.SetProjection(self.projection_string)
        quantity = eumQuantity.Create(EUMType.Bathymetry, EUMUnit.meter)
        builder.SetEumQuantity(quantity)
        newMesh = builder.CreateMesh()
        newMesh.Write(outfilename)



from __future__ import annotations
from pathlib import Path
from typing import Any
from mikecore.DfsuFile import DfsuFile, DfsuFileType

from ._dfsu import Dfsu2DH
from ._layered import Dfsu2DV, Dfsu3D
from ._spectral import DfsuSpectral


DFSU_MAPPING = {
    # DfsuFileType.Dfsu1D: None,
    DfsuFileType.Dfsu2D: Dfsu2DH,
    # DfsuFileType.DfsuVerticalColumn: None,
    DfsuFileType.DfsuVerticalProfileSigma: Dfsu2DV,
    DfsuFileType.DfsuVerticalProfileSigmaZ: Dfsu2DV,
    DfsuFileType.Dfsu3DSigma: Dfsu3D,
    DfsuFileType.Dfsu3DSigmaZ: Dfsu3D,
    DfsuFileType.DfsuSpectral0D: DfsuSpectral,
    DfsuFileType.DfsuSpectral1D: DfsuSpectral,
    DfsuFileType.DfsuSpectral2D: DfsuSpectral,
}


def dfsu(filename: str | Path) -> Any:
    filename = str(filename)
    dfs = DfsuFile.Open(filename)
    type = DfsuFileType(dfs.DfsuFileType)
    dfs.Close()

    klass = DFSU_MAPPING.get(type)

    if klass is None:
        raise ValueError(f"Unsupported dfsu type: {type}")

    return klass(filename)


class Dfsu:
    """Factory class for dfsu files."""

    def __new__(self, filename: str | Path) -> Any:
        return dfsu(filename)



from __future__ import annotations
from functools import cached_property
from pathlib import Path

from typing import Any, Iterable, Literal, Sequence
import warnings

from matplotlib.axes import Axes
import numpy as np
from mikecore.DfsuFile import DfsuFileType


from ._FM_geometry import GeometryFM2D, _GeometryFM, _GeometryFMPlotter
from ._geometry import GeometryPoint3D

from ._FM_utils import _plot_vertical_profile, BoundaryPolygons

from ._utils import _relative_cumulative_distance

Layer = Literal["all", "bottom", "top"]


class _GeometryFMLayered(_GeometryFM):
    def __init__(
        self,
        *,
        node_coordinates: np.ndarray,
        element_table: np.ndarray | list[Sequence[int]] | list[np.ndarray],
        codes: np.ndarray | None = None,
        projection: str = "LONG/LAT",
        dfsu_type: DfsuFileType = DfsuFileType.Dfsu3DSigma,
        element_ids: np.ndarray | None = None,
        node_ids: np.ndarray | None = None,
        n_layers: int = 1,  # at least 1 layer
        n_sigma: int | None = None,
        validate: bool = True,
        reindex: bool = False,
    ) -> None:
        super().__init__(
            node_coordinates=node_coordinates,
            element_table=element_table,
            codes=codes,
            projection=projection,
            dfsu_type=dfsu_type,
            element_ids=element_ids,
            node_ids=node_ids,
            validate=validate,
            reindex=reindex,
        )

        self._n_layers = n_layers
        self._n_sigma = n_sigma if n_sigma is not None else n_layers

        # Lazy properties
        self._bot_elems: np.ndarray | None = None
        self._e2_e3_table: np.ndarray | None = None
        self._2d_ids: np.ndarray | None = None
        self._layer_ids: np.ndarray | None = None

    def __repr__(self) -> str:
        return (
            f"Flexible Mesh Geometry: {self._type.name}\n"
            f"number of nodes: {self.n_nodes}\n"
            f"number of elements: {self.n_elements}\n"
            f"number of layers: {self.n_layers}\n"
            f"number of sigma layers: {self.n_sigma_layers}\n"
            f"projection: {self.projection_string}"
        )

    @cached_property
    def geometry2d(self) -> GeometryFM2D:
        return self.to_2d_geometry()

    def isel(
        self, idx: Sequence[int] | np.ndarray, keepdims: bool = False, **kwargs: Any
    ) -> GeometryFM3D | GeometryPoint3D | GeometryFM2D | GeometryFMVerticalColumn:
        return self.elements_to_geometry(elements=idx, keepdims=keepdims)

    def elements_to_geometry(
        self,
        elements: int | Sequence[int] | np.ndarray,
        node_layers: Layer = "all",
        keepdims: bool = False,
    ) -> GeometryFM3D | GeometryPoint3D | GeometryFM2D | GeometryFMVerticalColumn:
        sel_elements: list[int]

        if isinstance(elements, (int, np.integer)):
            sel_elements = [elements]
        else:
            sel_elements = list(elements)
        if len(sel_elements) == 1 and not keepdims:
            x, y, z = self.element_coordinates[sel_elements.pop(), :]

            return GeometryPoint3D(x=x, y=y, z=z, projection=self.projection)

        # create new geometry
        new_type = self._type

        layers_used = self.layer_ids[sel_elements]
        unique_layer_ids = np.unique(layers_used)
        n_layers = len(unique_layer_ids)

        if n_layers > 1:
            bottom: Layer = "bottom"
            elem_bot = self.get_layer_elements(layers=bottom)
            if np.all(np.isin(sel_elements, elem_bot)):
                n_layers = 1

        if (
            self._type == DfsuFileType.Dfsu3DSigma
            or self._type == DfsuFileType.Dfsu3DSigmaZ
        ) and n_layers == 1:
            new_type = DfsuFileType.Dfsu2D

        if n_layers == 1 and node_layers == "all":
            node_layers = "bottom"

        # extract information for selected elements
        if n_layers == 1:
            elem2d = self.elem2d_ids[sel_elements]
            geom2d = self.geometry2d
            node_ids, elem_tbl = geom2d._get_nodes_and_table_for_elements(elem2d)
            assert len(elem_tbl[0]) <= 4, "Not a 2D element"
            node_coords = geom2d.node_coordinates[node_ids]
            codes = geom2d.codes[node_ids]
            elem_ids = self._element_ids[elem2d]
        else:
            node_ids, elem_tbl = self._get_nodes_and_table_for_elements(
                sel_elements, node_layers=node_layers
            )
            node_coords = self.node_coordinates[node_ids]
            codes = self.codes[node_ids]
            elem_ids = self._element_ids[sel_elements]

        if new_type == DfsuFileType.Dfsu2D:
            return GeometryFM2D(
                node_coordinates=node_coords,
                codes=codes,
                node_ids=node_ids,
                projection=self.projection_string,
                element_table=elem_tbl,
                element_ids=elem_ids,
                dfsu_type=DfsuFileType.Dfsu2D,
                reindex=True,
            )
        else:
            lowest_sigma = self.n_layers - self.n_sigma_layers
            n_sigma = sum(unique_layer_ids >= lowest_sigma)
            if n_layers == len(elem_tbl):
                return GeometryFMVerticalColumn(
                    node_coordinates=node_coords,
                    codes=codes,
                    node_ids=node_ids,
                    projection=self.projection_string,
                    element_table=elem_tbl,
                    element_ids=elem_ids,
                    dfsu_type=self._type,
                    reindex=True,
                    n_layers=n_layers,
                    n_sigma=n_sigma,
                )
            else:
                klass = self.__class__
                return klass(  # type: ignore
                    node_coordinates=node_coords,
                    codes=codes,
                    node_ids=node_ids,
                    projection=self.projection_string,
                    element_table=elem_tbl,
                    element_ids=elem_ids,
                    dfsu_type=self._type,
                    reindex=True,
                    n_layers=n_layers,
                    n_sigma=n_sigma,
                )

    @cached_property
    def element_coordinates(self) -> np.ndarray:
        """Center coordinates of each element."""
        return self._calc_element_coordinates(maxnodes=8)

    def _get_nodes_and_table_for_elements(
        self,
        elements: Sequence[int] | np.ndarray,
        node_layers: Layer = "all",
    ) -> tuple[Any, Any]:
        """list of nodes and element table for a list of elements.

        Parameters
        ----------
        elements : np.array(int)
            array of element ids
        node_layers : str, optional
            for 3D files 'all', 'bottom' or 'top' nodes
            of each element, by default 'all'

        Returns
        -------
        np.array(int)
            array of node ids (unique)
        list(list(int))
            element table with a list of nodes for each element

        """
        elem_tbl = np.empty(len(elements), dtype=np.dtype("O"))
        if (node_layers == "all") or self.is_2d:
            for j, eid in enumerate(elements):
                elem_tbl[j] = np.asarray(self.element_table[eid])

        else:
            # 3D => 2D
            for j, eid in enumerate(elements):
                elem_nodes = np.asarray(self.element_table[eid])
                nn = len(elem_nodes)
                halfn = int(nn / 2)
                if node_layers == "bottom":
                    elem_nodes = elem_nodes[:halfn]
                if node_layers == "top":
                    elem_nodes = elem_nodes[halfn:]
                elem_tbl[j] = elem_nodes

        nodes = np.unique(np.hstack(elem_tbl))  # type: ignore
        return nodes, elem_tbl

    def to_2d_geometry(self) -> GeometryFM2D:
        """extract 2d geometry from 3d geometry.

        Returns
        -------
        GeometryFM2D
            2d geometry

        """
        # extract information for selected elements
        match self._type:
            case DfsuFileType.Dfsu3DSigmaZ:
                elem_ids = self.top_elements
            case DfsuFileType.Dfsu3DSigma:
                elem_ids = self.bottom_elements
            case _:
                raise NotImplementedError(
                    f"Conversion to 2D not implemented for {self._type}"
                )

        node_ids, elem_tbl = self._get_nodes_and_table_for_elements(
            elem_ids, node_layers="bottom"
        )
        node_coords = self.node_coordinates[node_ids]
        codes = self._codes[node_ids]

        geom = GeometryFM2D(
            node_coordinates=node_coords,
            codes=codes,
            node_ids=node_ids,
            projection=self.projection_string,
            element_table=elem_tbl,
            element_ids=elem_ids,
            validate=False,
            dfsu_type=DfsuFileType.Dfsu2D,
            reindex=True,
        )

        # TODO do this before creating the geometry

        # TODO extract method
        # Fix z-coordinate for sigma-z:
        if self._type == DfsuFileType.Dfsu3DSigmaZ:
            zn = geom.node_coordinates[:, 2].copy()
            for j, elem_nodes in enumerate(geom.element_table):
                elem_nodes3d = self.element_table[self.bottom_elements[j]]
                for jn in range(len(elem_nodes)):
                    znj_3d = self.node_coordinates[elem_nodes3d[jn], 2]
                    zn[elem_nodes[jn]] = min(zn[elem_nodes[jn]], znj_3d)
            geom.node_coordinates[:, 2] = zn

        return geom

    @cached_property
    def n_elements(self) -> int:
        """Number of 3d elements."""
        return len(self.element_table)

    @property
    def n_nodes(self) -> int:
        return len(self.node_coordinates)

    @property
    def is_2d(self) -> bool:
        return False

    @property
    def ndim(self) -> int:
        return 3

    @property
    def is_layered(self) -> bool:
        return True

    @cached_property
    def layer_ids(self) -> np.ndarray:
        """The layer number (0=bottom, 1, 2, ...) for each 3d element."""
        if self._layer_ids is None:
            res = self._get_2d_to_3d_association()
            self._e2_e3_table = res[0]
            self._2d_ids = res[1]
            self._layer_ids = res[2]
        return self._layer_ids

    @property
    def n_layers(self) -> int:
        """Maximum number of layers."""
        return self._n_layers

    @property
    def n_sigma_layers(self) -> int:
        """Number of sigma layers."""
        return self._n_sigma

    @property
    def n_z_layers(self) -> int:
        """Maximum number of z-layers."""
        return self.n_layers - self.n_sigma_layers

    @cached_property
    def top_elements(self) -> np.ndarray:
        """List of 3d element ids of surface layer."""
        # note: if subset of elements is selected then this cannot be done!

        # fast path if no z-layers
        if self.n_z_layers == 0:
            return np.arange(
                start=self.n_sigma_layers - 1,
                stop=self.n_elements,
                step=self.n_sigma_layers,
            )
        else:
            # slow path
            return self._find_top_layer_elements(self.element_table)

    def _elements_in_area(
        self, area: Sequence[tuple[float, float]] | Sequence[float]
    ) -> np.ndarray:
        """Find element ids of elements inside area."""
        idx2d = self.geometry2d._elements_in_area(area)
        if len(idx2d) > 0:
            return np.hstack(self.e2_e3_table[idx2d])
        else:
            return np.array([], dtype=int)

    @staticmethod
    def _find_top_layer_elements(elementTable: np.ndarray) -> np.ndarray:
        """Find element indices (zero based) of the elements being the upper-most element
        in its column.
        Each column is identified by matching node id numbers. For 3D elements the
        last half of the node numbers of the bottom element must match the first half
        of the node numbers in the top element. For 2D vertical elements the order of
        the node numbers in the bottom element (last half number of nodes) are reversed
        compared to those in the top element (first half number of nodes).
        To find the number of elements in each column, assuming the result
        is stored in res:
        For the first column it is res[0]+1.
        For the i'th column, it is res[i]-res[i-1].
        :returns: A list of element indices of top layer elements.
        """
        topLayerElments = []

        # Find top layer elements by matching the number numers of the last half of elmt i
        # with the first half of element i+1.
        # Elements always start from the bottom, and the element of one columne are following
        # each other in the element table.
        for i in range(len(elementTable) - 1):
            elmt1 = elementTable[i]
            elmt2 = elementTable[i + 1]

            if len(elmt1) != len(elmt2):
                # elements with different number of nodes can not be on top of each other,
                # so elmt2 must be another column, and elmt1 must be a top element
                topLayerElments.append(i)
                continue

            # Number of nodes in a 2D element
            elmt2DSize = len(elmt1) // 2

            for j in range(elmt2DSize):
                if elmt2DSize > 2:
                    if elmt1[j + elmt2DSize] != elmt2[j]:
                        # At least one node number did not match
                        # so elmt2 must be another column, and elmt1 must be a top element
                        topLayerElments.append(i)
                        break
                else:
                    # for 2D vertical profiles the nodes in the element on the
                    # top is in reverse order of those in the bottom.
                    if elmt1[j + elmt2DSize] != elmt2[(elmt2DSize - 1) - j]:
                        # At least one node number did not match
                        # so elmt2 must be another column, and elmt1 must be a top element
                        topLayerElments.append(i)
                        break

        # The last element will always be a top layer element
        topLayerElments.append(len(elementTable) - 1)

        return np.array(topLayerElments, dtype=np.int32)

    @cached_property
    def n_layers_per_column(self) -> np.ndarray:
        """List of number of layers for each column."""
        top_elems = self.top_elements
        n = len(top_elems)
        tmp = top_elems.copy()
        tmp[0] = -1
        tmp[1:n] = top_elems[0 : (n - 1)]
        n_layers_column = top_elems - tmp
        return n_layers_column

    @cached_property
    def bottom_elements(self) -> np.ndarray:
        """List of 3d element ids of bottom layer."""
        return self.top_elements - self.n_layers_per_column + 1

    def get_layer_elements(self, layers: int | Layer | Sequence[int]) -> np.ndarray:
        """3d element ids for one (or more) specific layer(s).

        Parameters
        ----------
        layers : int or list(int)
            layer between 0 (bottom) and n_layers-1 (top)
            (can also be negative counting from -1 at the top layer)

        Returns
        -------
        np.array(int)
            element ids

        """
        if isinstance(layers, str):
            if layers == "top":
                return self.top_elements
            elif layers == "bottom":
                return self.bottom_elements
            else:
                raise ValueError(
                    f"layers '{layers}' not recognized ('top', 'bottom' or integer)"
                )

        if not np.isscalar(layers):
            assert isinstance(layers, Iterable)
            elem_ids = []
            for layer in layers:
                assert isinstance(layer, int)
                elem_ids.append(self.get_layer_elements(layer))
            elem_ids = np.concatenate(elem_ids, axis=0)
            return np.sort(elem_ids)

        n_lay = self.n_layers

        assert isinstance(layers, int)
        if layers < (-n_lay) or layers >= n_lay:
            raise Exception(
                f"Layer {layers!r} not allowed; must be between -{n_lay} and {n_lay - 1}"
            )

        if layers < 0:
            layers = layers + n_lay

        return self._element_ids[self.layer_ids == layers]

    @property
    def e2_e3_table(self) -> np.ndarray:
        """The 2d-to-3d element connectivity table for a 3d object."""
        # e2_e3, 2d_ids and layer_ids are all set at the same time

        if self._e2_e3_table is None:
            res = self._get_2d_to_3d_association()
            self._e2_e3_table = res[0]
            self._2d_ids = res[1]
            self._layer_ids = res[2]
        return self._e2_e3_table

    @property
    def elem2d_ids(self) -> np.ndarray:
        """The associated 2d element id for each 3d element."""
        if self._2d_ids is None:
            res = self._get_2d_to_3d_association()
            self._e2_e3_table = res[0]
            self._2d_ids = res[1]
            self._layer_ids = res[2]
        return self._2d_ids

    def _get_2d_to_3d_association(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        e2_to_e3 = []  # for each 2d element: the corresponding 3d element ids from bot to top
        index2d = []  # for each 3d element: the associated 2d element id
        layerid = []  # for each 3d element: the associated layer number
        n2d = len(self.top_elements)
        topid = self.top_elements
        botid = self.bottom_elements
        # layer_ids = 0, 1, 2...
        global_layer_ids = np.arange(self.n_layers)
        for j in range(n2d):
            col = np.arange(botid[j], topid[j] + 1)

            e2_to_e3.append(col)
            for _ in col:
                index2d.append(j)

            n_local_layers = len(col)
            local_layers = global_layer_ids[-n_local_layers:]
            for ll in local_layers:
                layerid.append(ll)

        e2_to_e3 = np.array(e2_to_e3, dtype=object)
        index2d = np.array(index2d)
        layerid = np.array(layerid)
        return e2_to_e3, index2d, layerid

    def _z_idx_in_column(self, e3_col: np.ndarray, z: np.ndarray) -> np.ndarray:
        dz = self._dz[e3_col]
        z_col = self.element_coordinates[e3_col, 2]
        z_face = np.append(z_col - dz / 2, z_col[-1] + dz[-1] / 2)
        if z < z_face[0] or z > z_face[-1]:
            xy = tuple(self.element_coordinates[e3_col[0], :2])
            raise ValueError(
                f"z value '{z}' is outside water column [{z_face[0]},{z_face[-1]}] in point x,y={xy}"
            )
        idx = np.searchsorted(z_face, z) - 1
        return idx

    def _find_elem3d_from_elem2d(
        self, elem2d: int | np.ndarray, z: np.ndarray | float
    ) -> np.ndarray:
        """Find 3d element ids from 2d element ids and z-values."""
        # TODO: coordinate with _find_3d_from_2d_points()

        elem2d = [elem2d] if np.isscalar(elem2d) else elem2d
        elem2d = np.asarray(elem2d)
        z_vec = np.full(elem2d.shape, fill_value=z) if np.isscalar(z) else z

        elem3d = np.full_like(elem2d, fill_value=-1)
        for j, e2 in enumerate(elem2d):
            idx_3d = np.hstack(self.e2_e3_table[e2])
            elem3d[j] = idx_3d[self._z_idx_in_column(idx_3d, z_vec[j])]  # type: ignore

            # z_col = self.element_coordinates[idx_3d, 2]
            # elem3d[j] = (np.abs(z_col - z_vec[j])).argmin()  # nearest
        return elem3d

    # def _find_3d_from_2d_points(self, elem2d, z=None, layer=None):

    #     was_scalar = np.isscalar(elem2d)
    #     if was_scalar:
    #         elem2d = np.array([elem2d])
    #     else:
    #         orig_shape = elem2d.shape
    #         elem2d = np.reshape(elem2d, (elem2d.size,))

    #     if (layer is None) and (z is None):
    #         # return top element
    #         idx = self.top_elements[elem2d]  # TODO: return whole column instead

    #     elif layer is None:
    #         idx = np.zeros_like(elem2d)
    #         if np.isscalar(z):
    #             z = z * np.ones_like(elem2d, dtype=float)
    #         elem3d = self.e2_e3_table[elem2d]
    #         for j, row in enumerate(elem3d):
    #             zc = self.element_coordinates[row, 2]
    #             d3d = np.abs(z[j] - zc)
    #             idx[j] = row[d3d.argsort()[0]]

    #     elif z is None:
    #         if 0 <= layer <= self.n_z_layers - 1:
    #             idx = np.zeros_like(elem2d)
    #             elem3d = self.e2_e3_table[elem2d]
    #             for j, row in enumerate(elem3d):
    #                 try:
    #                     layer_ids = self.layer_ids[row]
    #                     id = row[list(layer_ids).index(layer)]
    #                     idx[j] = id
    #                 except IndexError:
    #                     raise IndexError(
    #                         f"Layer {layer} not present for 2d element {elem2d[j]}"
    #                     )
    #         else:
    #             # sigma layer
    #             idx = self.get_layer_elements(layer)[elem2d]

    #     else:
    #         raise ValueError("layer and z cannot both be supplied!")

    #     if was_scalar:
    #         idx = idx[0]
    #     else:
    #         idx = np.reshape(idx, orig_shape)

    #     return idx

    @cached_property
    def _dz(self) -> np.ndarray:
        """Height of each 3d element (using static zn information)."""
        return self._calc_dz()

    def _calc_dz(self) -> np.ndarray:
        """Height of 3d elements using static or dynamic zn information."""
        element_table = self.element_table
        n_elements = len(element_table)

        zn = self.node_coordinates[:, 2]
        zn_is_2d = len(zn.shape) == 2
        shape = (zn.shape[0], n_elements) if zn_is_2d else (n_elements,)
        dz = np.full(shape=shape, fill_value=np.nan)

        # static zn
        for j in range(n_elements):
            nodes = element_table[j]
            halfn = int(len(nodes) / 2)
            z_bot = np.mean(zn[nodes[:halfn]])
            z_top = np.mean(zn[nodes[halfn:]])
            dz[j] = z_top - z_bot

        return dz


class GeometryFM3D(_GeometryFMLayered):
    """Flexible 3d mesh geometry."""

    def __init__(
        self,
        *,
        node_coordinates: np.ndarray,
        element_table: np.ndarray | list[Sequence[int]] | list[np.ndarray],
        codes: np.ndarray | None = None,
        projection: str = "LONG/LAT",
        dfsu_type: DfsuFileType = DfsuFileType.Dfsu3DSigma,
        element_ids: np.ndarray | None = None,
        node_ids: np.ndarray | None = None,
        n_layers: int = 1,  # at least 1 layer
        n_sigma: int | None = None,
        validate: bool = True,
        reindex: bool = False,
    ) -> None:
        super().__init__(
            node_coordinates=node_coordinates,
            element_table=element_table,
            codes=codes,
            projection=projection,
            dfsu_type=dfsu_type,
            element_ids=element_ids,
            node_ids=node_ids,
            n_layers=n_layers,
            n_sigma=n_sigma,
            validate=validate,
            reindex=reindex,
        )
        self.plot = _GeometryFMPlotter(self)

    @property
    def boundary_polylines(self) -> BoundaryPolygons:
        warnings.warn(
            "boundary_polylines is renamed to boundary_polygons", FutureWarning
        )
        return self.geometry2d.boundary_polylines

    @property
    def boundary_polygons(self) -> BoundaryPolygons:
        return self.geometry2d.boundary_polylines

    def contains(self, points: np.ndarray) -> np.ndarray:
        return self.geometry2d.contains(points)

    def to_mesh(self, outfilename: str | Path) -> None:
        return self.geometry2d.to_mesh(outfilename)

    def find_index(
        self,
        x: float | None = None,
        y: float | None = None,
        z: float | None = None,
        coords: np.ndarray | None = None,
        area: tuple[float, float, float, float] | None = None,
        layers: int | Layer | Sequence[int] | None = None,
    ) -> np.ndarray:
        if layers is not None:
            idx = self.get_layer_elements(layers)
        else:
            idx = self._element_ids

        if (
            (coords is not None)
            or (x is not None)
            or (y is not None)
            or (z is not None)
        ):
            if area is not None:
                raise ValueError(
                    "Coordinates and area cannot be provided at the same time!"
                )
            if coords is not None:
                coords = np.atleast_2d(coords)
                xy = coords[:, :2]  # type: ignore
                z = coords[:, 2] if coords.shape[1] == 3 else None  # type: ignore
            else:
                xy = np.vstack((x, y)).T  # type: ignore
            idx_2d = self.geometry2d._find_element_2d(coords=xy)
            assert len(idx_2d) == len(xy)
            if z is None:
                idx_3d = np.hstack(self.e2_e3_table[idx_2d])
            else:
                idx_3d = self._find_elem3d_from_elem2d(idx_2d, z)
            idx = np.intersect1d(idx, idx_3d).astype(int)
        elif area is not None:
            idx_area = self._elements_in_area(area)
            idx = np.intersect1d(idx, idx_area)
        elif layers is None:
            raise ValueError(
                "At least one selection argument (x,y,z,coords,area,layers) needs to be provided!"
            )
        return idx


class GeometryFMVerticalProfile(_GeometryFMLayered):
    """Flexible mesh 2d vertical profile geometry."""

    def __init__(
        self,
        node_coordinates: np.ndarray,
        element_table: np.ndarray | list[Sequence[int]] | list[np.ndarray],
        codes: np.ndarray | None = None,
        projection: str = "LONG/LAT",
        dfsu_type: DfsuFileType = DfsuFileType.Dfsu3DSigma,
        element_ids: np.ndarray | None = None,
        node_ids: np.ndarray | None = None,
        n_layers: int = 1,  # at least 1 layer
        n_sigma: int | None = None,
        validate: bool = True,
        reindex: bool = False,
    ) -> None:
        super().__init__(
            node_coordinates=node_coordinates,
            element_table=element_table,
            codes=codes,
            projection=projection,
            dfsu_type=dfsu_type,
            element_ids=element_ids,
            node_ids=node_ids,
            n_layers=n_layers,
            n_sigma=n_sigma,
            validate=validate,
            reindex=reindex,
        )
        self.plot = _GeometryFMVerticalProfilePlotter(self)

    @cached_property
    def relative_element_distance(self) -> np.ndarray:
        ec = self.element_coordinates
        nc0 = self.node_coordinates[0, :2]
        return _relative_cumulative_distance(ec, nc0, is_geo=self.is_geo)

    def get_nearest_relative_distance(self, coords: tuple[float, float]) -> float:
        """For a point near a transect, find the nearest relative distance
        for showing position on transect plot.

        Parameters
        ----------
        coords : [float, float]
            x,y-coordinate of point

        Returns
        -------
        float
            relative distance in meters from start of transect

        """
        xe = self.element_coordinates[:, 0]
        ye = self.element_coordinates[:, 1]
        dd2 = np.square(xe - coords[0]) + np.square(ye - coords[1])
        idx = np.argmin(dd2)
        return self.relative_element_distance[idx]

    def find_index(
        self,
        x: float | None = None,
        y: float | None = None,
        z: float | None = None,
        coords: np.ndarray | None = None,
        layers: int | Sequence[int] | Layer | None = None,
    ) -> np.ndarray:
        if layers is not None:
            idx = self.get_layer_elements(layers)
        else:
            idx = self._element_ids

        # select in space
        if (
            (coords is not None)
            or (x is not None)
            or (y is not None)
            or (z is not None)
        ):
            if coords is not None:
                coords = np.atleast_2d(coords)
                xy = coords[:, :2]  # type: ignore
                z = coords[:, 2] if coords.shape[1] == 3 else None  # type: ignore
            else:
                xy = np.vstack((x, y)).T  # type: ignore

            idx_2d = self._find_nearest_element_2d(coords=xy)

            if z is None:
                idx_3d = np.hstack(self.e2_e3_table[idx_2d])
            else:
                idx_3d = self._find_elem3d_from_elem2d(idx_2d, z)
            idx = np.intersect1d(idx, idx_3d)

        return idx

    def _find_nearest_element_2d(self, coords: np.ndarray) -> np.ndarray:
        ec2d = self.element_coordinates[self.top_elements, :2]
        xe, ye = ec2d[:, 0], ec2d[:, 1]
        coords = np.atleast_2d(coords)
        idx = np.zeros(len(coords), dtype=int)
        for j, pt in enumerate(coords):
            x, y = pt[0:2]
            idx[j] = np.argmin((xe - x) ** 2 + (ye - y) ** 2)
        return idx


class GeometryFMVerticalColumn(GeometryFM3D):
    "A 3d geometry with consisting of a single vertical column."

    # TODO: add plotter

    def calc_ze(self, zn: np.ndarray | None = None) -> np.ndarray:
        if zn is None:
            zn = self.node_coordinates[:, 2]
        return self._calc_z_using_idx(zn, self._idx_e)

    def calc_zf(self, zn: np.ndarray | None = None) -> np.ndarray:
        if zn is None:
            zn = self.node_coordinates[:, 2]
        return self._calc_z_using_idx(zn, self._idx_f)

    def _calc_zee(self, zn: np.ndarray | None = None) -> np.ndarray:
        ze = self.calc_ze(zn)
        zf = self.calc_zf(zn)
        if ze.ndim == 1:
            zee = np.insert(ze, 0, zf[0])
            return np.append(zee, zf[-1])
        else:
            return np.hstack(
                (zf[:, 0].reshape((-1, 1)), ze, zf[:, -1].reshape((-1, 1)))
            )

    def _interp_values(
        self, zn: np.ndarray, data: np.ndarray, z: np.ndarray
    ) -> np.ndarray:
        """Interpolate to other z values, allow linear extrapolation."""
        from scipy.interpolate import interp1d  # type: ignore

        opt = {"kind": "linear", "bounds_error": False, "fill_value": "extrapolate"}

        ze = self.calc_ze(zn)
        dati = np.zeros_like(z)
        if zn.ndim == 1:
            dati = interp1d(ze, data, **opt)(z)
        elif zn.ndim == 2:
            for j in range(zn.shape[0]):
                dati[j, :] = interp1d(ze[j, :], data[j, :], **opt)(z[j, :])
        return dati

    @property
    def _idx_f(self) -> np.ndarray:
        nnodes_half = int(len(self.element_table[0]) / 2)
        n_vfaces = self.n_elements + 1
        idx_f = np.zeros((n_vfaces, nnodes_half), dtype=int)
        idx_e = self._idx_e
        idx_f[: self.n_elements, :] = idx_e[:, :nnodes_half]
        idx_f[self.n_elements, :] = idx_e[-1, nnodes_half:]
        return idx_f

    @property
    def _idx_e(self) -> np.ndarray:
        nnodes_per_elem = len(self.element_table[0])
        idx_e = np.zeros((self.n_elements, nnodes_per_elem), dtype=int)

        for j in range(self.n_elements):
            nodes = self.element_table[j]
            for i in range(nnodes_per_elem):
                idx_e[j, i] = nodes[i]

        return idx_e

    def _calc_z_using_idx(self, zn: np.ndarray, idx: np.ndarray) -> np.ndarray:
        return zn[..., idx].mean(axis=-1)


class _GeometryFMVerticalProfilePlotter:
    def __init__(self, geometry: "GeometryFMVerticalProfile") -> None:
        self.g = geometry

    def __call__(
        self,
        ax: Axes | None = None,
        figsize: tuple[float, float] | None = None,
        **kwargs: Any,
    ) -> Axes:
        import matplotlib.pyplot as plt

        if ax is None:
            _, ax = plt.subplots(figsize=figsize)
        x = self.g.node_coordinates[:, 0]
        y = self.g.node_coordinates[:, 1]
        ax.plot(x, y, **kwargs)
        return ax

    def mesh(self, title: str = "Mesh", edge_color: str = "0.5", **kwargs: Any) -> Axes:
        v = np.full_like(self.g.element_coordinates[:, 0], np.nan)
        return _plot_vertical_profile(
            node_coordinates=self.g.node_coordinates,
            element_table=self.g.element_table,
            values=v,
            is_geo=self.g.is_geo,
            title=title,
            add_colorbar=False,
            edge_color=edge_color,
            cmin=0.0,
            cmax=1.0,
            **kwargs,
        )



from dataclasses import dataclass
from typing import Any, Literal, Sequence

from numpy.typing import NDArray
from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.collections import PatchCollection
from matplotlib.colors import Colormap, Normalize
from matplotlib.figure import Figure
from matplotlib.tri import Triangulation
import numpy as np
from scipy.sparse import csr_matrix

from ._utils import _relative_cumulative_distance


MESH_COL = "0.95"
MESH_COL_DARK = "0.6"


@dataclass
class Polygon:
    xy: NDArray

    @property
    def area(self) -> float:
        return (
            np.dot(self.xy[:, 1], np.roll(self.xy[:, 0], 1))
            - np.dot(self.xy[:, 0], np.roll(self.xy[:, 1], 1))
        ) * 0.5


@dataclass
class BoundaryPolygons:
    exteriors: list[Polygon]
    interiors: list[Polygon]

    @property
    def lines(self) -> list[Polygon]:
        return self.exteriors + self.interiors

    def contains(self, points: np.ndarray) -> np.ndarray:
        """Test if a list of points are contained by mesh.

        Parameters
        ----------
        points : array-like n-by-2
            x,y-coordinates of n points to be tested

        Returns
        -------
        bool array
            True for points inside, False otherwise

        """
        import matplotlib.path as mp  # type: ignore

        exterior = self.exteriors[0]
        cnts = mp.Path(exterior.xy).contains_points(points)

        if len(self.exteriors) > 1:
            # in case of several dis-joint outer domains
            for exterior in self.exteriors[1:]:
                in_domain = mp.Path(exterior.xy).contains_points(points)
                cnts = np.logical_or(cnts, in_domain)

        # subtract any holes
        for interior in self.interiors:
            in_hole = mp.Path(interior.xy).contains_points(points)
            cnts = np.logical_and(cnts, ~in_hole)

        return cnts


def _plot_map(
    node_coordinates: np.ndarray,
    element_table: np.ndarray,
    element_coordinates: np.ndarray,
    boundary_polylines: list[Polygon],
    projection: str = "",
    z: np.ndarray | None = None,
    plot_type: Literal[
        "patch", "mesh_only", "shaded", "contour", "contourf", "outline_only"
    ] = "patch",
    title: str | None = None,
    label: str | None = None,
    cmap: Colormap | None = None,
    vmin: float | None = None,
    vmax: float | None = None,
    levels: int | Sequence[float] | None = None,
    n_refinements: int = 0,
    show_mesh: bool = False,
    show_outline: bool = True,
    figsize: tuple[float, float] | None = None,
    ax: Axes | None = None,
    add_colorbar: bool = True,
) -> Axes:
    """Plot unstructured data and/or mesh, mesh outline.

    Parameters
    ----------
    node_coordinates: np.array
        node coordinates
    element_table: np.array
        element table
    element_coordinates: np.array
        element coordinates
    boundary_polylines: BoundaryPolylines,
        boundary polylines
    projection: str, optional
        projection type, default: ""
    z: np.array or a Dataset with a single item, optional
        value for each element to plot, default bathymetry
    plot_type: str, optional
        type of plot: 'patch' (default), 'mesh_only', 'shaded',
        'contour', 'contourf' or 'outline_only'
    title: str, optional
        axes title
    label: str, optional
        colorbar label (or title if contour plot)
    cmap: matplotlib.cm.cmap, optional
        colormap, default viridis
    vmin: real, optional
        lower bound of values to be shown on plot, default:None
    vmax: real, optional
        upper bound of values to be shown on plot, default:None
    levels: int, list(float), optional
        for contour plots: how many levels, default:10
        or a list of discrete levels e.g. [3.0, 4.5, 6.0]
    show_mesh: bool, optional
        should the mesh be shown on the plot? default=True
    show_outline: bool, optional
        should domain outline be shown on the plot? default=True
    n_refinements: int, optional
        for 'shaded' and 'contour' plots (and if show_mesh=False)
        do this number of mesh refinements for smoother plotting
    figsize: (float, float), optional
        specify size of figure
    ax: matplotlib.axes, optional
        Adding to existing axis, instead of creating new fig
    add_colorbar: bool
        Add colorbar to plot, default True

    Returns
    -------
    <matplotlib.axes>

    Examples
    --------
    >>> dfs = Dfsu("HD2D.dfsu")
    >>> dfs.plot() # bathymetry
    >>> ds = dfs.read(items="Surface elevation", time=0)
    >>> ds.shape
    (1, 884)
    >>> ds.n_items
    1
    >>> dfs.plot(z=ds) # plot surface elevation

    """
    import matplotlib.pyplot as plt
    import matplotlib

    VALID_PLOT_TYPES = (
        "mesh_only",
        "outline_only",
        "contour",
        "contourf",
        "patch",
        "shaded",
    )
    if plot_type not in VALID_PLOT_TYPES:
        ok_list = ", ".join(VALID_PLOT_TYPES)
        raise Exception(f"plot_type {plot_type} unknown! ({ok_list})")

    cmap = cmap or matplotlib.colormaps["viridis"]

    nc = node_coordinates
    ec = element_coordinates

    if ((vmin is not None) or (vmax is not None)) and (
        levels is not None and not np.isscalar(levels)
    ):
        raise ValueError(
            "vmin/vmax cannot be provided together with non-integer levels"
        )

    # plot in existing or new axes?
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)

    # set aspect ratio
    __set_aspect_ratio(ax, nc, projection)
    _set_xy_label_by_projection(ax, projection)

    if plot_type == "outline_only":
        __plot_outline_only(ax, boundary_polylines)
        return ax

    if plot_type == "mesh_only":
        __plot_mesh_only(ax, nc, element_table)
        return ax

    # At this point we are sure that we are plotting some data, at least bathymetry
    if z is None:
        z = ec[:, 2]
        label = label or "Bathymetry (m)"

    assert len(z) == ec.shape[0]

    label = label or ""

    vmin, vmax, cmap, cmap_norm, cmap_ScMappable, levels = __set_colormap_levels(
        cmap, vmin, vmax, levels, z
    )
    cbar_extend = __cbar_extend(z, vmin, vmax)

    if plot_type == "patch":
        fig_obj: Any = __plot_patch(
            ax, nc, element_table, show_mesh, cmap, cmap_norm, z, vmin, vmax
        )

    else:
        # do node-based triangular plot
        mesh_linewidth = 0.0
        if show_mesh and __is_tri_only(element_table):
            mesh_linewidth = 0.4
            n_refinements = 0
        triang, zn = __get_tris(nc, element_table, ec, z, n_refinements)

        if plot_type == "shaded":
            ax.triplot(triang, lw=mesh_linewidth, color=MESH_COL)
            fig_obj = ax.tripcolor(
                triang,
                zn,
                edgecolors="face",
                cmap=cmap,
                vmin=vmin,
                vmax=vmax,
                linewidths=0.3,
                shading="gouraud",
            )

        elif plot_type == "contour":
            ax.triplot(triang, lw=mesh_linewidth, color=MESH_COL_DARK)
            fig_obj = ax.tricontour(
                triang,
                zn,
                levels=levels,
                linewidths=[1.2],
                cmap=cmap,
                norm=cmap_norm,
            )
            ax.clabel(fig_obj, fmt="%1.2f", inline=1, fontsize=9)
            ax.set_title(label)
            add_colorbar = False

        elif plot_type == "contourf":
            ax.triplot(triang, lw=mesh_linewidth, color=MESH_COL)
            fig_obj = ax.tricontourf(
                triang,
                zn,
                levels=levels,
                cmap=cmap,
                norm=cmap_norm,
                extend=cbar_extend,
                vmin=vmin,
                vmax=vmax,
            )

        if show_mesh and (not __is_tri_only(element_table)):
            __add_non_tri_mesh(ax, nc, element_table, plot_type)

    if show_outline:
        __add_outline(ax, boundary_polylines)

    if add_colorbar:
        __add_colorbar(ax, cmap_ScMappable, fig_obj, label, levels, cbar_extend)

    __set_plot_limits(ax, nc)

    if title:
        ax.set_title(title)

    return ax


def __set_colormap_levels(
    cmap: Colormap | str,
    vmin: float | None,
    vmax: float | None,
    levels: int | Sequence[float] | np.ndarray | None,
    z: np.ndarray,
) -> tuple[float, float, Colormap, Normalize, ScalarMappable, np.ndarray]:
    """Set colormap, levels, vmin, vmax, and cmap_norm.

    Parameters
    ----------
    cmap : str or matplotlib.colors.Colormap
        colormap name or colormap object
    vmin : float
        minimum value for colorbar
    vmax : float
        maximum value for colorbar
    levels : int or list of float
        number of levels or list of levels
    z : array of float
        data to be plotted

    Returns
    -------
    vmin : float
        minimum value for colorbar
    vmax : float
        maximum value for colorbar
    cmap : matplotlib.colors.Colormap
        colormap object
    cmap_norm : matplotlib.colors.Normalize
        colormap normalization object
    cmap_ScMappable : matplotlib.cm.ScalarMappable
        colormap object
    levels : list of float
        list of levels

    """
    import matplotlib
    import matplotlib.cm as cm
    import matplotlib.colors as mplc

    vmin = vmin or np.nanmin(z)
    vmax = vmax or np.nanmax(z)

    if vmin == vmax:
        vmin = vmin - 0.1
        vmax = vmin + 0.2

    cmap_norm = None
    cmap_ScMappable = None
    if levels is not None:
        if np.isscalar(levels):
            n_levels = levels
            levels = np.linspace(vmin, vmax, n_levels)  # type: ignore
        else:
            n_levels = len(levels)  # type: ignore
            vmin = min(levels)  # type: ignore
            vmax = max(levels)  # type: ignore

        levels = np.array(levels)

        if isinstance(cmap, str):
            cmap = matplotlib.colormaps[cmap]
        cmap_norm = mplc.BoundaryNorm(levels, cmap.N)
        cmap_ScMappable = cm.ScalarMappable(cmap=cmap, norm=cmap_norm)

    if levels is None:
        levels = np.linspace(vmin, vmax, 10)

    return vmin, vmax, cmap, cmap_norm, cmap_ScMappable, levels  # type: ignore


def __set_plot_limits(ax: Axes, nc: np.ndarray) -> None:
    """Set default plot limits.

    Override with matplotlib ax.set_xlim, ax.set_ylim
    """
    xmin, xmax = nc[:, 0].min(), nc[:, 0].max()
    ymin, ymax = nc[:, 1].min(), nc[:, 1].max()

    xybuf = 6e-3 * (xmax - xmin)
    ax.set_xlim(xmin - xybuf, xmax + xybuf)
    ax.set_ylim(ymin - xybuf, ymax + xybuf)


def __plot_mesh_only(ax: Axes, nc: np.ndarray, element_table: np.ndarray) -> None:
    """plot mesh only (no data).

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        axes object
    nc : array of float
        node coordinates
    element_table : array of int
        element table

    Returns
    -------
    matplotlib.axes.Axes
        axes object

    """
    from matplotlib.collections import PatchCollection

    patches = _to_polygons(nc, element_table)
    fig_obj = PatchCollection(
        patches, edgecolor=MESH_COL_DARK, facecolor="none", linewidths=0.3
    )
    ax.add_collection(fig_obj)


def __plot_outline_only(ax: Axes, boundary_polylines: list[Polygon]) -> Axes:
    """plot outline only (no data).

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        axes object
    boundary_polylines : list[PolyLine]
        boundary polylines

    Returns
    -------
    matplotlib.axes.Axes
        axes object

    """
    __add_outline(ax, boundary_polylines)
    return ax


def __plot_patch(
    ax: Axes,
    nc: np.ndarray,
    element_table: np.ndarray,
    show_mesh: bool,
    cmap: Colormap,
    cmap_norm: Normalize,
    z: np.ndarray,
    vmin: float,
    vmax: float,
) -> PatchCollection:
    """plot patch with data from z.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        axes object
    nc : array of float
        node coordinates
    element_table : array of int
        element table
    show_mesh : bool
        include mesh polygons
    cmap : str or matplotlib.colors.Colormap
        colormap name or colormap object
    cmap_norm : matplotlib.colors.Normalize
        colormap normalization object
    z : array of float
        data to be plotted
    vmin : float
        minimum value for colorbar
    vmax : float
        maximum value for colorbar

    Returns
    -------
    matplotlib.axes.Axes
        axes object

    """
    patches = _to_polygons(nc, element_table)

    if show_mesh:
        edgecolor = MESH_COL
        linewidth = 0.4
    else:
        edgecolor = "face"
        linewidth = None

    fig_obj = PatchCollection(
        patches,
        cmap=cmap,
        norm=cmap_norm,
        edgecolor=edgecolor,
        linewidths=linewidth,
    )

    fig_obj.set_array(z)
    fig_obj.set_clim(vmin, vmax)
    ax.add_collection(fig_obj)

    return fig_obj


def __get_tris(
    nc: np.ndarray,
    element_table: np.ndarray,
    ec: np.ndarray,
    z: np.ndarray,
    n_refinements: int,
) -> tuple[Triangulation, np.ndarray]:
    """get triangulation object and node-centered data.

    Parameters
    ----------
    nc : array of float
        node coordinates
    element_table : array of int
        element table
    ec : array of int
        element coordinates
    z : array of float
        data to be plotted
    n_refinements : int
        number of refinements

    Returns
    -------
    matplotlib.tri.Triangulation and node-centered data

    """
    import matplotlib.tri as tri

    elem_table, ec, z = __create_tri_only_element_table(nc, element_table, ec, data=z)
    triang = tri.Triangulation(nc[:, 0], nc[:, 1], elem_table)

    zn = _get_node_centered_data(nc, elem_table, ec, z)

    if n_refinements > 0:
        # TODO: refinements doesn't seem to work for 3d files?
        refiner = tri.UniformTriRefiner(triang)
        triang, zn = refiner.refine_field(zn, subdiv=n_refinements)

    return triang, zn


def __add_colorbar(
    ax: Axes,
    cmap_ScMappable: ScalarMappable,
    fig_obj: Figure,
    label: str,
    levels: np.ndarray,
    cbar_extend: str,
) -> None:
    """add colorbar to axes.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        axes object
    cmap_ScMappable : matplotlib.cm.ScalarMappable
        colormap object
    fig_obj : matplotlib.figure.Figure
        figure object
    label : str
        colorbar label
    levels : array of float
        colorbar levels
    cbar_extend : str
        extend colorbar beyond min/max values

    Returns
    -------
    None

    """
    from mpl_toolkits.axes_grid1 import make_axes_locatable  # type: ignore
    import matplotlib.pyplot as plt

    cax = make_axes_locatable(ax).append_axes("right", size="5%", pad=0.05)
    cmap_sm = cmap_ScMappable if cmap_ScMappable else fig_obj

    plt.colorbar(
        cmap_sm,  # type: ignore
        label=label,
        cax=cax,
        ticks=levels,
        boundaries=levels,
        extend=cbar_extend,
    )


def __set_aspect_ratio(ax: Axes, nc: np.ndarray, projection: str) -> None:
    """set aspect ratio.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        axes object
    nc : array of float
        node coordinates
    projection : str
        projection type

    Returns
    -------
    None

    """
    is_geo = projection == "LONG/LAT"
    if is_geo:
        mean_lat = np.mean(nc[:, 1])
        ax.set_aspect(1.0 / np.cos(np.pi * mean_lat / 180))
    else:
        ax.set_aspect("equal")


def __add_non_tri_mesh(
    ax: Axes, nc: np.ndarray, element_table: np.ndarray, plot_type: str
) -> None:
    """add non-triangular mesh to axes.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        axes object
    nc : array of float
        node coordinates
    element_table : array of int
        element table
    plot_type : str
        type of plot

    Returns
    -------
    None

    """
    # if mesh is not tri only, we need to add it manually on top
    from matplotlib.collections import PatchCollection

    patches = _to_polygons(nc, element_table)
    mesh_linewidth = 0.4
    if plot_type == "contour":
        mesh_col = MESH_COL_DARK
    else:
        mesh_col = MESH_COL
    p = PatchCollection(
        patches,
        edgecolor=mesh_col,
        facecolor="none",
        linewidths=mesh_linewidth,
    )
    ax.add_collection(p)


def __add_outline(ax: Axes, boundary_polylines: list[Polygon]) -> None:
    """add outline to axes.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        axes object
    boundary_polylines: list[PolyLine]
        boundary polylines

    Returns
    -------
    None

    """
    for line in boundary_polylines:
        ax.plot(*line.xy.T, color="0.4", linewidth=1.2)


def _set_xy_label_by_projection(ax: Axes, projection: str) -> None:
    if (not projection) or projection == "NON-UTM":
        ax.set_xlabel("x [m]")
        ax.set_ylabel("y [m]")
    elif projection == "LONG/LAT":
        ax.set_xlabel("Longitude [degrees]")
        ax.set_ylabel("Latitude [degrees]")
    else:
        ax.set_xlabel("Easting [m]")
        ax.set_ylabel("Northing [m]")


def __is_tri_only(element_table: np.ndarray) -> bool:
    return max([len(el) for el in element_table]) == 3


def _to_polygons(node_coordinates: np.ndarray, element_table: np.ndarray) -> list[Any]:
    """generate matplotlib polygons from element table for plotting.

    Returns
    -------
    list(matplotlib.patches.Polygon)
        list of polygons for plotting

    """
    from matplotlib.patches import Polygon

    polygons = []

    for j in range(len(element_table)):
        nodes = element_table[j]
        pcoords = np.empty([len(nodes), 2])
        for i in range(len(nodes)):
            nidx = nodes[i]
            pcoords[i, :] = node_coordinates[nidx, 0:2]

        polygon = Polygon(pcoords, closed=True)
        polygons.append(polygon)
    return polygons


def _create_node_element_matrix(
    element_table: np.ndarray, num_nodes: int
) -> csr_matrix:
    """Creates a sparse node-element connectivity matrix from a given element table.

    Parameters
    ----------
    element_table : np.array
        The element table (A 2D array where each row represents an element and each
        column corresponds to a node index involved in the element.)
    num_nodes : int
        The total number of nodes in the mesh.

    Returns
    -------
    scipy.sparse.csr_matrix
        A sparse matrix of shape (num_nodes, number of elements), where the entry
        (i, j) is 1 if node i is part of element j, and 0 otherwise.

    """
    row_ind = element_table.ravel()
    col_ind = np.repeat(np.arange(element_table.shape[0]), element_table.shape[1])
    data = np.ones(len(row_ind), dtype=int)
    connectivity_matrix = csr_matrix(
        (data, (row_ind, col_ind)), shape=(num_nodes, element_table.shape[0])
    )
    return connectivity_matrix


def _get_node_centered_data(
    node_coordinates: np.ndarray,
    element_table: np.ndarray,
    element_coordinates: np.ndarray,
    data: np.ndarray,
    extrapolate: bool = True,
) -> np.ndarray:
    """convert cell-centered data to node-centered by pseudo-laplacian method.

    Parameters
    ----------
    node_coordinates: np.array(float)
        node coordinates
    element_table: np.array[int]
        element table
    element_coordinates: np.array(float)
        element coordinates
    data : np.array(float)
        cell-centered data
    extrapolate : bool, optional
        allow the method to extrapolate, default:True

    Returns
    -------
    np.array(float)
        node-centered data

    """
    nc = node_coordinates
    elem_table, ec, data = __create_tri_only_element_table(
        nc, element_table, element_coordinates, data
    )
    connectivity_matrix = _create_node_element_matrix(elem_table, nc.shape[0])

    node_centered_data = np.zeros(shape=nc.shape[0])
    for n in range(connectivity_matrix.shape[0]):
        item = connectivity_matrix.getrow(n).indices
        I = ec[item][:, :2] - nc[n][:2]
        I2 = (I**2).sum(axis=0)
        Ixy = (I[:, 0] * I[:, 1]).sum(axis=0)
        lamb = I2[0] * I2[1] - Ixy**2
        omega = np.zeros(1)
        if lamb > 1e-10 * (I2[0] * I2[1]):
            # Standard case - Pseudo
            lambda_x = (Ixy * I[:, 1] - I2[1] * I[:, 0]) / lamb
            lambda_y = (Ixy * I[:, 0] - I2[0] * I[:, 1]) / lamb
            omega = 1.0 + lambda_x * I[:, 0] + lambda_y * I[:, 1]
            if not extrapolate:
                omega[np.where(omega > 2)] = 2
                omega[np.where(omega < 0)] = 0
        if omega.sum() > 0:
            node_centered_data[n] = np.sum(omega * data[item]) / np.sum(omega)
        else:
            # We did not succeed using pseudo laplace procedure, use inverse distance instead
            InvDis = [
                1 / np.hypot(case[0], case[1]) for case in ec[item][:, :2] - nc[n][:2]
            ]
            node_centered_data[n] = np.sum(InvDis * data[item]) / np.sum(InvDis)

    return node_centered_data


def __create_tri_only_element_table(
    node_coordinates: np.ndarray,
    element_table: np.ndarray,
    element_coordinates: np.ndarray,
    data: np.ndarray,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Convert quad/tri mesh to pure tri-mesh."""
    if __is_tri_only(element_table):
        # already tri-only? just convert to 2d array
        return np.stack(element_table), element_coordinates, data  # type: ignore

    ec = element_coordinates.copy()

    elem_table = [list(element_table[i]) for i in range(len(element_table))]
    tmp_elmnt_nodes = elem_table.copy()
    for el, item in enumerate(tmp_elmnt_nodes):
        if len(item) == 4:
            elem_table.pop(el)  # remove quad element

            # insert two new tri elements in table
            elem_table.insert(el, item[:3])
            tri2_nodes = [item[i] for i in [2, 3, 0]]
            elem_table.append(tri2_nodes)

            # new center coordinates for new tri-elements
            ec[el] = node_coordinates[item[:3]].mean(axis=1)
            tri2_ec = node_coordinates[tri2_nodes].mean(axis=1)
            ec = np.append(ec, tri2_ec.reshape(1, -1), axis=0)

            # use same data in two new tri elements
            data = np.append(data, data[el])

    return np.asarray(elem_table), ec, data


def __cbar_extend(
    calc_data: np.ndarray | None, vmin: float | None, vmax: float | None
) -> str:
    if calc_data is None:
        return "neither"
    extend_min = calc_data.min() < vmin if vmin is not None else False
    extend_max = calc_data.max() > vmax if vmax is not None else False
    if extend_min and extend_max:
        extend = "both"
    elif extend_min:
        extend = "min"
    elif extend_max:
        extend = "max"
    else:
        extend = "neither"
    return extend


def _plot_vertical_profile(
    node_coordinates: np.ndarray,
    element_table: np.ndarray,
    values: np.ndarray,
    zn: np.ndarray | None = None,
    is_geo: bool = False,
    cmin: float | None = None,
    cmax: float | None = None,
    label: str = "",
    add_colorbar: bool = True,
    **kwargs: Any,
) -> Axes:
    """Plot unstructured vertical profile.

    Parameters
    ----------
    node_coordinates: np.array
        node coordinates
    element_table: np.array
        element table
    values: np.array
        value for each element to plot
    zn: np.array, optional
        dynamic vertical node positions,
        default: use static vertical positions
    is_geo: bool, optional
        are coordinates geographical (for calculating
        relative distance in meters), default: False
    cmin: real, optional
        lower bound of values to be shown on plot, default:None
    cmax: real, optional
        upper bound of values to be shown on plot, default:None
    title: str, optional
        axes title
    label: str, optional
        colorbar label
    cmap: matplotlib.cm.cmap, optional
        colormap, default viridis
    edge_color: str, optional
        color of mesh lines, default: None
    add_colorbar: bool, optional
        Add colorbar to plot, default True
    figsize: (float, float), optional
        specify size of figure
    ax: matplotlib.axes, optional
        Adding to existing axis, instead of creating new fig
    **kwargs: Any
        Additional keyword arguments

    Returns
    -------
    <matplotlib.axes>

    """
    import matplotlib.pyplot as plt
    from matplotlib.collections import PolyCollection

    nc = node_coordinates
    s_coordinate = _relative_cumulative_distance(nc, is_geo=is_geo)
    z_coordinate = nc[:, 2] if zn is None else zn

    elements = _Get_2DVertical_elements(element_table)

    # plot in existing or new axes?
    if "ax" in kwargs:
        ax = kwargs["ax"]
    else:
        figsize = None
        if "figsize" in kwargs:
            figsize = kwargs["figsize"]
        _, ax = plt.subplots(figsize=figsize)

    sz = np.c_[s_coordinate, z_coordinate]
    verts = sz[elements]

    if "cmap" in kwargs:
        cmap = kwargs["cmap"]
    else:
        cmap = "jet"
    pc = PolyCollection(verts, cmap=cmap)

    if cmin is None:
        cmin = np.nanmin(values)
    if cmax is None:
        cmax = np.nanmax(values)
    pc.set_clim(cmin, cmax)

    if add_colorbar:
        plt.colorbar(pc, ax=ax, label=label, orientation="vertical")
    pc.set_array(values)

    if "edge_color" in kwargs:
        edge_color = kwargs["edge_color"]
    else:
        edge_color = None
    pc.set_edgecolor(edge_color)

    ax.add_collection(pc)
    ax.autoscale()
    ax.set_xlabel("relative distance [m]")
    ax.set_ylabel("z [m]")

    if "title" in kwargs:
        ax.set_title(kwargs["title"])

    return ax


def _Get_2DVertical_elements(element_table: np.ndarray) -> np.ndarray:
    # if (type == DfsuFileType.DfsuVerticalProfileSigmaZ) or (
    #     type == DfsuFileType.DfsuVerticalProfileSigma
    # ):
    elements = [list(element_table[i]) for i in range(len(list(element_table)))]
    return np.asarray(elements)



# content of conftest.py
import pytest
import numpy
import pandas
import mikeio


@pytest.fixture(autouse=True)
def add_np(doctest_namespace):
    doctest_namespace["np"] = numpy

@pytest.fixture(autouse=True)
def add_pd(doctest_namespace):
    doctest_namespace["pd"] = pandas


@pytest.fixture(autouse=True)
def add_mikeio(doctest_namespace):
    doctest_namespace["mikeio"] = mikeio


from __future__ import annotations
from pathlib import Path
from collections.abc import Sequence
from typing import Any, Callable

import numpy as np
import pandas as pd

from .dataset import Dataset
from .dfs import Dfs0
from .eum import ItemInfo, EUMUnit, EUMType
from .spatial import GeometryFM2D


def _extract_track(
    *,
    deletevalue: float,
    start_time: pd.Timestamp,
    end_time: pd.Timestamp,
    timestep: float,
    geometry: GeometryFM2D,
    track: str | Path | Dataset | pd.DataFrame,
    items: Sequence[ItemInfo],
    item_numbers: Sequence[int],
    time_steps: Sequence[int],
    n_elements: int,
    method: str,
    dtype: Any,  # TODO DTypeLike?
    data_read_func: Callable[[int, int], tuple[np.ndarray, float]],
) -> Dataset:
    if not isinstance(geometry, GeometryFM2D):
        raise NotImplementedError("Only implemented for 2d flexible mesh geometries")

    n_items = len(item_numbers)

    match track:
        case str():
            times, coords = _get_track_data_from_file(track)
        case Path():
            times, coords = _get_track_data_from_file(str(track))
        case Dataset():
            times, coords = _get_track_data_from_dataset(track)
        case pd.DataFrame():
            times, coords = _get_track_data_from_dataframe(track)
        case _:
            raise ValueError(
                "track must be a file name, a Dataset or a pandas DataFrame"
            )

    assert isinstance(
        times, pd.DatetimeIndex
    ), "The index must be a pandas.DatetimeIndex"
    assert times.is_monotonic_increasing, "The time index must be monotonic increasing. Consider df.sort_index() before passing to extract_track()."

    data_list = [coords[:, 0], coords[:, 1]]  # lon,lat
    for item in range(n_items):
        # Initialize an empty data block
        data = np.empty(shape=(len(times)), dtype=dtype)
        data[:] = np.nan
        data_list.append(data)

    if geometry.is_geo:
        lon = coords[:, 0]
        lon[lon < -180] = lon[lon < -180] + 360
        lon[lon >= 180] = lon[lon >= 180] - 360
        coords[:, 0] = lon

    # track end (relative time)
    t_rel = (times - end_time).total_seconds()
    # largest idx for which (times - self.end_time)<=0
    tmp = np.where(t_rel <= 0)[0]
    if len(tmp) == 0:
        raise ValueError("No time overlap!")
    i_end = tmp[-1]

    # track time relative to start
    t_rel = (times - start_time).total_seconds()
    tmp = np.where(t_rel >= 0)[0]
    if len(tmp) == 0:
        raise ValueError("No time overlap!")
    i_start = tmp[0]  # smallest idx for which t_rel>=0

    dfsu_step = int(np.floor(t_rel[i_start] / timestep))  # first step

    # spatial interpolation
    n_pts = 1 if method == "nearest" else 5
    elem_ids, weights = geometry.get_2d_interpolant(
        coords[i_start : (i_end + 1)], n_nearest=n_pts
    )

    # initialize arrays
    d1: np.ndarray = np.ndarray(shape=(n_items, n_elements), dtype=dtype)
    d2: np.ndarray = np.ndarray(shape=(n_items, n_elements), dtype=dtype)
    t1 = 0.0
    t2 = 0.0

    # very first dfsu time step
    step = time_steps[dfsu_step]
    for i, item in enumerate(item_numbers):
        d, t2 = data_read_func(item, step)
        t2 = t2 - 1e-10  # TODO what is this operation doing?
        d[d == deletevalue] = np.nan
        d2[i, :] = d

    def is_EOF(step: int) -> bool:
        return step >= len(time_steps)

    # loop over track points
    for i_interp, t in enumerate(range(i_start, i_end + 1)):
        t_rel[t]  # time of point relative to dfsu start

        read_next = t_rel[t] > t2

        while read_next and not is_EOF(dfsu_step + 1):
            dfsu_step = dfsu_step + 1

            # swap new to old
            d1, d2 = d2, d1
            t1, t2 = t2, t1

            step = time_steps[dfsu_step]
            for i, item in enumerate(item_numbers):
                d, t2 = data_read_func(item, step)
                d[d == deletevalue] = np.nan
                d2[i, :] = d

            read_next = t_rel[t] > t2

        if read_next and is_EOF(dfsu_step):
            # cannot read next - no more timesteps
            continue

        w = (t_rel[t] - t1) / timestep  # time-weight
        eid = elem_ids[i_interp]
        if np.any(eid > 0):
            dati = (1 - w) * np.dot(d1[:, eid], weights[i_interp])
            dati = dati + w * np.dot(d2[:, eid], weights[i_interp])
        else:
            dati = np.empty(shape=n_items, dtype=dtype)
            dati[:] = np.nan

        for item in range(n_items):
            data_list[item + 2][t] = dati[item]

    if geometry.is_geo:
        items_out = [
            ItemInfo("Longitude", EUMType.Latitude_longitude, EUMUnit.degree),
            ItemInfo("Latitude", EUMType.Latitude_longitude, EUMUnit.degree),
        ]
    else:
        items_out = [
            ItemInfo("x", EUMType.Geographical_coordinate, EUMUnit.meter),
            ItemInfo("y", EUMType.Geographical_coordinate, EUMUnit.meter),
        ]

    for item_info in items:
        items_out.append(item_info)

    return Dataset(data_list, times, items_out)


def _get_track_data_from_dataset(track: Dataset) -> tuple[pd.DatetimeIndex, np.ndarray]:
    times = track.time
    coords = np.zeros(shape=(len(times), 2))
    coords[:, 0] = track[0].to_numpy().copy()
    coords[:, 1] = track[1].to_numpy().copy()
    return times, coords


def _get_track_data_from_dataframe(
    track: pd.DataFrame,
) -> tuple[pd.DatetimeIndex, np.ndarray]:
    times = track.index
    coords = track.iloc[:, 0:2].to_numpy(copy=True)
    return times, coords


def _get_track_data_from_file(track: str) -> tuple[pd.DatetimeIndex, np.ndarray]:
    filename = track
    path = Path(filename)
    if not path.exists():
        raise FileNotFoundError(f"{filename} does not exist")

    ext = path.suffix.lower()
    match ext:
        case ".dfs0":
            df = Dfs0(filename).to_dataframe()
        case ".csv":
            df = pd.read_csv(filename, index_col=0, parse_dates=True)
            df.index = pd.DatetimeIndex(df.index)
        case _:
            raise ValueError(f"{ext} files not supported (dfs0, csv)")

    times = df.index
    coords = df.iloc[:, 0:2].to_numpy(copy=True)

    return times, coords



from __future__ import annotations
from typing import Any, Sequence


import numpy as np
from mikecore.DfsuFile import DfsuFileType


from ._geometry import _Geometry

from ._FM_geometry import GeometryFM2D


class GeometryFMPointSpectrum(_Geometry):
    """Flexible mesh point spectrum."""

    def __init__(
        self,
        frequencies: np.ndarray | None = None,
        directions: np.ndarray | None = None,
        x: float | None = None,
        y: float | None = None,
    ) -> None:
        super().__init__()
        self.n_nodes = 0
        self.n_elements = 0
        self.is_2d = False
        self.is_spectral = True

        self._frequencies = frequencies
        self._directions = directions
        self.x = x
        self.y = y

    @property
    def default_dims(self) -> tuple[str, ...]:
        if self.directions is None:
            return ("frequency",)
        else:
            return ("direction", "frequency")

    @property
    def is_layered(self) -> bool:
        return False

    def __repr__(self) -> str:
        txt = f"Point Spectrum Geometry(frequency:{self.n_frequencies}, direction:{self.n_directions}"
        if self.x is not None:
            txt = txt + f", x:{self.x:.5f}, y:{self.y:.5f}"
        return txt + ")"

    @property
    def ndim(self) -> int:
        # TODO: 0, 1 or 2 ?
        return 0

    @property
    def n_frequencies(self) -> int:
        """Number of frequencies."""
        return 0 if self.frequencies is None else len(self.frequencies)

    @property
    def frequencies(self) -> np.ndarray | None:
        """Frequency axis."""
        return self._frequencies

    @property
    def n_directions(self) -> int:
        """Number of directions."""
        return 0 if self.directions is None else len(self.directions)

    @property
    def directions(self) -> np.ndarray | None:
        """Directional axis."""
        return self._directions


class _GeometryFMSpectrum(GeometryFM2D):
    def __init__(
        self,
        node_coordinates: np.ndarray,
        element_table: Any,
        codes: np.ndarray | None = None,
        projection: str = "LONG/LAT",
        dfsu_type: DfsuFileType | None = None,
        element_ids: np.ndarray | None = None,
        node_ids: np.ndarray | None = None,
        validate: bool = True,
        frequencies: np.ndarray | None = None,
        directions: np.ndarray | None = None,
        reindex: bool = False,
    ) -> None:
        super().__init__(
            node_coordinates=node_coordinates,
            element_table=element_table,
            codes=codes,
            projection=projection,
            dfsu_type=dfsu_type,
            element_ids=element_ids,
            node_ids=node_ids,
            validate=validate,
            reindex=reindex,
        )

        self._frequencies = frequencies
        self._directions = directions

    @property
    def n_frequencies(self) -> int:
        """Number of frequencies."""
        return 0 if self.frequencies is None else len(self.frequencies)

    @property
    def frequencies(self) -> np.ndarray | None:
        """Frequency axis."""
        return self._frequencies

    @property
    def n_directions(self) -> int:
        """Number of directions."""
        return 0 if self.directions is None else len(self.directions)

    @property
    def directions(self) -> np.ndarray | None:
        """Directional axis."""
        return self._directions


# TODO reconsider inheritance to avoid overriding method signature
class GeometryFMAreaSpectrum(_GeometryFMSpectrum):
    """Flexible mesh area spectrum geometry."""

    def isel(  # type: ignore
        self, idx: Sequence[int], **kwargs: Any
    ) -> "GeometryFMPointSpectrum" | "GeometryFMAreaSpectrum":
        return self.elements_to_geometry(elements=idx)

    def elements_to_geometry(  # type: ignore
        self, elements: Sequence[int], keepdims: bool = False
    ) -> "GeometryFMPointSpectrum" | "GeometryFMAreaSpectrum":
        """export a selection of elements to new flexible file geometry
        Parameters.
        ----------
        elements : list(int)
            list of element ids
        keepdims: bool
            Not used
        Returns
        -------
        GeometryFMAreaSpectrum or GeometryFMPointSpectrum
            which can be used for further extraction or saved to file

        """
        elements = np.atleast_1d(elements)  # type: ignore
        if len(elements) == 1:
            coords = self.element_coordinates[elements[0], :]
            return GeometryFMPointSpectrum(
                frequencies=self._frequencies,
                directions=self._directions,
                x=coords[0],
                y=coords[1],
            )

        node_ids, elem_tbl = self._get_nodes_and_table_for_elements(elements)
        node_coords = self.node_coordinates[node_ids]
        codes = self.codes[node_ids]

        geom = GeometryFMAreaSpectrum(
            node_coordinates=node_coords,
            codes=codes,
            node_ids=node_ids,
            projection=self.projection_string,
            element_table=elem_tbl,
            element_ids=self.element_ids[elements],
            frequencies=self._frequencies,
            directions=self._directions,
            reindex=True,
        )
        geom._type = self._type
        return geom


# TODO this inherits indirectly from GeometryFM2D, which is not ideal
class GeometryFMLineSpectrum(_GeometryFMSpectrum):
    """Flexible mesh line spectrum geometry."""

    def isel(  # type: ignore
        self, idx: Sequence[int], axis: str = "node"
    ) -> GeometryFMPointSpectrum | GeometryFMLineSpectrum:
        return self._nodes_to_geometry(nodes=idx)

    def _nodes_to_geometry(  # type: ignore
        self, nodes: Sequence[int]
    ) -> GeometryFMPointSpectrum | GeometryFMLineSpectrum:
        """export a selection of nodes to new flexible file geometry
        Note: takes only the elements for which all nodes are selected
        Parameters.
        ----------
        nodes : list(int)
            list of node ids
        Returns
        -------
        GeometryFMPointSpectrum | GeometryFMLineSpectrum
            which can be used for further extraction or saved to file

        """
        nodes = np.atleast_1d(nodes)
        if len(nodes) == 1:
            coords = self.node_coordinates[nodes[0], :2]
            return GeometryFMPointSpectrum(
                frequencies=self._frequencies,
                directions=self._directions,
                x=coords[0],
                y=coords[1],
            )

        elements = []
        for j, el_nodes in enumerate(self.element_table):
            if np.all(np.isin(el_nodes, nodes)):
                elements.append(j)

        assert len(elements) > 0, "no elements found"

        node_ids, elem_tbl = self._get_nodes_and_table_for_elements(elements)
        node_coords = self.node_coordinates[node_ids]
        codes = self.codes[node_ids]

        geom = GeometryFMLineSpectrum(
            node_coordinates=node_coords,
            codes=codes,
            node_ids=node_ids,
            projection=self.projection_string,
            dfsu_type=self._type,
            element_table=elem_tbl,
            element_ids=self.element_ids[elements],
            frequencies=self._frequencies,
            directions=self._directions,
            reindex=True,
        )
        return geom



from __future__ import annotations
from pathlib import Path
from collections.abc import Sequence
from typing import Any

import numpy as np
import pandas as pd

from mikecore.DfsBuilder import DfsBuilder
from mikecore.DfsFactory import DfsFactory
from mikecore.DfsFile import DfsFile, DfsSimpleType
from mikecore.DfsFileFactory import DfsFileFactory
from mikecore.eum import eumQuantity, eumUnit
from mikecore.Projections import Cartography

from .. import __dfs_version__
from ..dataset import Dataset
from ._dfs import (
    _Dfs123,
    _get_item_info,
    _valid_item_numbers,
    _valid_timesteps,
    write_dfs_data,
)
from ..eum import TimeStepUnit
from ..spatial import Grid3D


def write_dfs3(filename: str | Path, ds: Dataset, title: str = "") -> None:
    dfs = _write_dfs3_header(filename, ds, title)
    write_dfs_data(dfs=dfs, ds=ds, n_spatial_dims=3)


def _write_dfs3_header(filename: str | Path, ds: Dataset, title: str) -> DfsFile:
    builder = DfsBuilder.Create(title, "mikeio", __dfs_version__)
    builder.SetDataType(0)

    geometry: Grid3D = ds.geometry

    factory = DfsFactory()
    builder.SetSpatialAxis(
        factory.CreateAxisEqD3(
            eumUnit.eumUmeter,
            geometry.nx,
            geometry.x[0],
            geometry.dx,
            geometry.ny,
            geometry.y[0],
            geometry.dy,
            geometry.nz,
            geometry.z[0],
            geometry.dz,
        )
    )
    origin = geometry.origin  # Origin in geographical coordinates
    orient = geometry.orientation

    if geometry.is_geo:
        proj = factory.CreateProjectionGeoOrigin(
            geometry.projection_string, *origin, orient
        )
    else:
        cart: Cartography = Cartography.CreateProjOrigin(
            geometry.projection_string, *origin, orient
        )
        proj = factory.CreateProjectionGeoOrigin(
            wktProjectionString=geometry.projection,
            lon0=cart.LonOrigin,
            lat0=cart.LatOrigin,
            orientation=cart.Orientation,
        )

    builder.SetGeographicalProjection(proj)

    timestep_unit = TimeStepUnit.SECOND
    dt = ds.timestep or 1.0  # It can not be None
    if ds.is_equidistant:
        time_axis = factory.CreateTemporalEqCalendarAxis(
            timestep_unit, ds.time[0], 0, dt
        )
    else:
        time_axis = factory.CreateTemporalNonEqCalendarAxis(timestep_unit, ds.time[0])
    builder.SetTemporalAxis(time_axis)

    for item in ds.items:
        builder.AddCreateDynamicItem(
            item.name,
            eumQuantity.Create(item.type, item.unit),
            DfsSimpleType.Float,
            item.data_value_type,
        )

    try:
        builder.CreateFile(str(filename))
    except IOError:
        print("cannot create dfs file: ", filename)

    return builder.GetFile()


class Dfs3(_Dfs123):
    """Class for reading/writing dfs3 files."""

    _ndim = 3

    def __init__(self, filename: str | Path):
        super().__init__(str(filename))

        # TODO
        self._x0 = 0.0
        self._y0 = 0.0
        self._z0 = 0.0

        self._read_dfs3_header()
        self._validate_no_orientation_in_geo()
        origin, orientation = self._origin_and_orientation_in_CRS()

        self._geometry = Grid3D(
            x0=self._x0,
            dx=self._dx,
            nx=self._nx,
            y0=self._y0,
            dy=self._dy,
            ny=self._ny,
            z0=self._z0,
            dz=self._dz,
            nz=self._nz,
            origin=origin,
            projection=self._projstr,
            orientation=orientation,
        )

    def _read_dfs3_header(self, read_x0y0z0: bool = False) -> None:
        self._dfs = DfsFileFactory.Dfs3FileOpen(self._filename)

        self._source = self._dfs

        if read_x0y0z0:
            self._x0 = self._dfs.SpatialAxis.X0
            self._y0 = self._dfs.SpatialAxis.Y0
            self._z0 = self._dfs.SpatialAxis.Z0

        self._dx = self._dfs.SpatialAxis.Dx
        self._dy = self._dfs.SpatialAxis.Dy
        self._dz = self._dfs.SpatialAxis.Dz
        self._nx = self._dfs.SpatialAxis.XCount
        self._ny = self._dfs.SpatialAxis.YCount
        self._nz = self._dfs.SpatialAxis.ZCount

    def read(
        self,
        *,
        items: str | int | Sequence[str | int] | None = None,
        time: int | str | slice | None = None,
        area: tuple[float, float, float, float] | None = None,
        layers: str | int | Sequence[int] | None = None,
        keepdims: bool = False,
        dtype: Any = np.float32,
    ) -> Dataset:
        """Read data from a dfs3 file.

        Parameters
        ---------
        items: list[int] or list[str], optional
            Read only selected items, by number (0-based), or by name
        time: int, str, datetime, pd.TimeStamp, sequence, slice or pd.DatetimeIndex, optional
            Read only selected time steps, by default None (=all)
        area: tuple[float, float, float, float], optional
            Read only data within the specified rectangular area (x0, x1, y0, y1)
        keepdims: bool, optional
            When reading a single time step or a single layer only,
            should the singleton dimension be kept
            in the returned Dataset? by default: False
        layers: int, str, list[int], optional
            Read only data for specific layers, by default None
        dtype: data-type, optional
            Define the dtype of the returned dataset (default = np.float32)

        Returns
        -------
        Dataset

        """
        if area is not None:
            raise NotImplementedError("area subsetting is not yet implemented for Dfs3")
        # NOTE:
        # if keepdims is not False:
        #    return NotImplementedError("keepdims is not yet implemented for Dfs3")

        dfs = DfsFileFactory.DfsGenericOpen(self._filename)

        item_numbers = _valid_item_numbers(dfs.ItemInfo, items)
        n_items = len(item_numbers)

        single_time_selected, time_steps = _valid_timesteps(
            dfs.FileInfo, time_steps=time
        )
        nt = len(time_steps) if not single_time_selected else 1

        nz = self.geometry.nz
        ny = self.geometry.ny
        nx = self.geometry.nx
        deleteValue = dfs.FileInfo.DeleteValueFloat

        data_list = []

        if layers == "top":
            layers = -1
        layers = None if layers is None else np.atleast_1d(layers)

        dims: tuple[str, ...]
        shape: tuple[int, ...]

        nzl = nz if layers is None else len(layers)
        if nzl == 1 and (not keepdims):
            geometry = self.geometry._geometry_for_layers([0])
            dims = ("time", "y", "x")
            shape = (nt, ny, nx)
        else:
            geometry = self.geometry._geometry_for_layers(layers, keepdims)  # type: ignore
            dims = ("time", "z", "y", "x")
            shape = (nt, nzl, ny, nx)

        for item in range(n_items):
            data: np.ndarray = np.ndarray(shape=shape, dtype=dtype)
            data_list.append(data)

        if single_time_selected and not keepdims:
            shape = shape[1:]
            dims = tuple([d for d in dims if d != "time"])

        t_seconds = np.zeros(nt, dtype=float)

        for i, it in enumerate(time_steps):
            for item in range(n_items):
                itemdata = dfs.ReadItemTimeStep(item_numbers[item] + 1, int(it))
                d = itemdata.Data

                d = d.reshape(nz, ny, nx)
                d[d == deleteValue] = np.nan

                if layers is None:
                    dd = d
                elif len(layers) == 1:
                    if layers[0] == "bottom":
                        dd = self._get_bottom_values(d)
                    else:
                        dd = d[layers[0], :, :]
                else:
                    dd = d[layers, :, :]

                if single_time_selected and not keepdims:
                    data_list[item] = dd
                else:
                    data_list[item][i, ...] = dd

            t_seconds[i] = itemdata.Time

        dfs.Close()

        time = pd.to_datetime(t_seconds, unit="s", origin=self.start_time)
        items = _get_item_info(dfs.ItemInfo, item_numbers)
        return Dataset(
            data_list,
            time=time,
            items=items,
            geometry=geometry,
            dims=dims,
            validate=False,
        )

    def append(self, ds: Dataset, validate: bool = True) -> None:
        """Append a Dataset to an existing dfs3 file.

        Parameters
        ----------
        ds: Dataset
            Dataset to append
        validate: bool, optional
            Validate that the dataset to append has the same geometry and items as the original file

        Notes
        -----
        The original file is modified.

        """
        if validate:
            if self.geometry != ds.geometry:
                raise ValueError("The geometry of the dataset to append does not match")

            for item_s, item_o in zip(ds.items, self.items):
                if item_s != item_o:
                    raise ValueError(
                        f"Item in dataset {item_s.name} does not match {item_o.name}"
                    )

        dfs = DfsFileFactory.Dfs3FileOpenAppend(str(self._filename))
        write_dfs_data(dfs=dfs, ds=ds, n_spatial_dims=3)
        self._n_timesteps = dfs.FileInfo.TimeAxis.NumberOfTimeSteps

    @staticmethod
    def _get_bottom_values(data: np.ndarray) -> np.ndarray:
        assert len(data.shape) == 3
        b = np.empty_like(data[0])
        b[:] = np.nan
        data = np.flipud(data)
        for layer in range(data.shape[0]):  # going from surface to bottom
            y = data[layer, ...]
            b[~np.isnan(y)] = y[~np.isnan(y)]

        return b

    @property
    def geometry(self) -> Grid3D:
        return self._geometry

    @property
    def dx(self) -> float:
        """Step size in x direction."""
        return self._dx

    @property
    def dy(self) -> float:
        """Step size in y direction."""
        return self._dy

    @property
    def dz(self) -> float:
        """Step size in y direction."""
        return self._dz

    @property
    def shape(self) -> tuple[int, int, int, int]:
        return (self._n_timesteps, self._nz, self._ny, self._nx)



from __future__ import annotations
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path
from collections.abc import Sequence
from typing import Any

try:
    # read version from installed package
    __version__ = version("mikeio")
except PackageNotFoundError:
    # package is not installed
    __version__ = "dev"

__dfs_version__: int = 220


from .dataset import DataArray, Dataset, from_pandas, from_polars
from .dfs import Dfs0, Dfs1, Dfs2, Dfs3
from .dfsu import Dfsu, Mesh, Dfsu2DH, Dfsu2DV, Dfsu3D, DfsuSpectral
from .eum import EUMType, EUMUnit, ItemInfo
from .pfs import PfsDocument, PfsSection, read_pfs

from .spatial import (
    Grid1D,
    Grid2D,
    Grid3D,
    GeometryFM2D,
    GeometryFM3D,
    GeometryFMVerticalProfile,
)


def read(
    filename: str | Path,
    *,
    items: str | int | Sequence[str | int] | None = None,
    time: int | str | slice | None = None,
    keepdims: bool = False,
    **kwargs: Any,
) -> Dataset:
    """Read all or a subset of the data from a dfs file.

    All dfs files can be subsetted with the *items* and *time* arguments. But
    the following file types also have the shown additional arguments:

    * Dfs2: area
    * Dfs3: layers
    * Dfsu-2d: (x,y), elements, area
    * Dfsu-layered: (xy,z), elements, area, layers

    Parameters
    ----------
    filename
        full path and file name to the dfs file.
    items: int, str, list[int] or list[str], optional
        Read only selected items, by number (0-based), or by name,
        by default None (=all)
    time: int, str, datetime, pd.TimeStamp, sequence, slice or pd.DatetimeIndex, optional
        Read only selected time steps, by default None (=all)
    keepdims: bool, optional
        When reading a single time step only, should the time-dimension be kept
        in the returned Dataset? by default: False
    x, y, z: float, optional
        Dfsu: Read only data for elements containing the (x,y)
        or (x,y,z) points(s), by default None
    area: (float, float, float, float), optional
        Dfs2/Dfsu: read only data within an area given by a bounding
        box of coordinates (left, lower, right, upper), by default None (=all)
    layers: int, str or sequence, optional
        Dfs3/Dfsu-layered: read only data from specific layers,
        by default None (=all layers)
    error_bad_data: bool, optional
            raise error if data is corrupt, by default True,
    fill_bad_data_value:
            fill value for to impute corrupt data, used in conjunction with error_bad_data=False
            default np.nan
    **kwargs: Any
        Additional keyword arguments

    Returns
    -------
    Dataset
        A Dataset with specification according to the file type

    See also
    --------
    mikeio.open - open a Dfs file and only read the header

    Examples
    --------
    >>> ds = mikeio.read("ts.dfs0")
    >>> ds = mikeio.read("ts.dfs0", items=0)
    >>> ds = mikeio.read("ts.dfs0", items="Temperature")
    >>> ds = mikeio.read("sw_points.dfs0, items="*Buoy 4*")
    >>> ds = mikeio.read("ts.dfs0", items=["u","v"], time="2016")
    >>> ds = mikeio.read("tide.dfs1", time="2018-5")
    >>> ds = mikeio.read("tide.dfs1", time=slice("2018-5-1","2018-6-1"))
    >>> ds = mikeio.read("tide.dfs1", items=[0,3,6], time=-1)
    >>> ds = mikeio.read("tide.dfs1", time=-1, keepdims=True)
    >>> ds = mikeio.read("era5.dfs2", area=(10,50,16,58))
    >>> ds = mikeio.read("HD2D.dfsu")
    >>> ds = mikeio.read("HD2D.dfsu", x=2.2, y=54.2)
    >>> ds = mikeio.read("HD2D.dfsu", elements=183)
    >>> ds = mikeio.read("HD2D.dfsu", elements=range(0,2000))
    >>> ds = mikeio.read("HD2D.dfsu", area=(10,50,16,58))
    >>> ds = mikeio.read("MT3D_sigma_z.dfsu", x=11.4, y=56.2)
    >>> ds = mikeio.read("MT3D_sigma_z.dfsu", x=11.4, y=56.2, z=-1.1)
    >>> ds = mikeio.read("MT3D_sigma_z.dfsu", elements=lst_of_elems)
    >>> ds = mikeio.read("MT3D_sigma_z.dfsu", layers="bottom")
    >>> ds = mikeio.read("MT3D_sigma_z.dfsu", layers=[-2,-1])
    >>> ds = mikeio.read("HD2D.dfsu", error_bad_data=False) # replace corrupt data with np.nan
    >>> ds = mikeio.read("HD2D.dfsu", error_bad_data=False, fill_bad_data_value=0.0) # replace corrupt data with 0.0

    """
    ext = Path(filename).suffix.lower()

    if "dfs" not in ext:
        raise ValueError(
            "mikeio.read() is only supported for dfs files. Use mikeio.open for mesh files."
        )

    dfs = open(filename)
    if isinstance(dfs, Mesh):
        raise ValueError("mikeio.read() is not supported for Mesh files")

    return dfs.read(items=items, time=time, keepdims=keepdims, **kwargs)


def open(
    filename: str | Path, **kwargs: Any
) -> Dfs0 | Dfs1 | Dfs2 | Dfs3 | Dfsu2DH | Dfsu2DV | Dfsu3D | DfsuSpectral | Mesh:
    """Open a dfs/mesh file (and read the header).

    The typical workflow for small dfs files is to read all data
    with *mikeio.read* instead of using this function. For big files, however,
    it can be convenient to open the file first with *dfs=mikeio.open(...)* to
    inspect it's content (items, time and shape) and then decide what to
    read using dfs.read(...)

    Parameters
    ----------
    filename
        full path and file name to the dfs file.
    type : str, optional
        Dfs2 only. Additional information about the file, e.g.
        "spectral" for spectral dfs2 files. By default: None.
    **kwargs: Any
        Additional keyword arguments, e.g. *type="spectral"*

    See also
    --------
    mikeio.read - read data from a dfs file

    Examples
    --------
    >>> dfs = mikeio.open("wl.dfs1")
    >>> dfs = mikeio.open("HD2D.dfsu")
    >>> ds = dfs.read(items="Salinity", time="2016-01")

    >>> dfs = mikeio.open("pt_spectra.dfs2", type="spectral")

    """
    ext = Path(filename).suffix.lower()[1:]

    READERS = {
        "dfs0": Dfs0,
        "dfs1": Dfs1,
        "dfs2": Dfs2,
        "dfs3": Dfs3,
        "dfsu": Dfsu,
        "mesh": Mesh,
    }

    if ext not in READERS:
        valid_formats = ", ".join(READERS.keys())
        raise Exception(
            f"{ext} is not a supported format for mikeio.open. Valid formats are {valid_formats}"
        )

    reader_klass = READERS[ext]

    return reader_klass(filename, **kwargs)


__all__ = [
    "DataArray",
    "Dataset",
    "Dfs0",
    "Dfs1",
    "Dfs2",
    "Dfs3",
    "Dfsu",
    "Dfsu2DH",
    "Dfsu2DV",
    "Dfsu3D",
    "DfsuSpectral",
    "Mesh",
    "EUMType",
    "EUMUnit",
    "ItemInfo",
    "Pfs",
    "PfsDocument",
    "PfsSection",
    "read_pfs",
    "Grid1D",
    "Grid2D",
    "Grid3D",
    "GeometryFM2D",
    "GeometryFM3D",
    "GeometryFMVerticalProfile",
    "read",
    "open",
    "from_pandas",
    "from_polars",
]
